import numpy as np

from .statement import Statement
from numeta.syntax.nodes import Node
from numeta.syntax.settings import settings
from .tools import get_shape_blocks
from numeta.array_shape import SCALAR, UNKNOWN


class VariableDeclaration(Statement):
    def __init__(self, variable, add_to_scope=False):
        super().__init__(add_to_scope=add_to_scope)
        self.variable = variable

    def extract_entities(self):
        yield from self.variable._ftype.extract_entities()

        if settings.array_lower_bound != 1:
            # HACK: Non stardard array lower bound so we have to shift it
            # and well need the integer kind
            yield from settings.DEFAULT_INTEGER.extract_entities()

        if self.variable._shape is not UNKNOWN:
            for element in self.variable._shape.dims:
                if isinstance(element, Node):
                    yield from element.extract_entities()

    def get_code_blocks(self):
        result = self.variable._ftype.get_code_blocks()

        if self.variable.allocatable:
            result += [", ", "allocatable"]
            result += [", ", "dimension"]
            result += ["("] + [":", ","] * (len(self.variable._shape.dims) - 1) + [":", ")"]
        elif self.variable.pointer:
            result += [", ", "pointer"]
            if self.variable._shape is not SCALAR:
                result += [", ", "dimension"]
                result += ["("] + [":", ","] * (len(self.variable._shape.dims) - 1) + [":", ")"]
        elif self.variable._shape is UNKNOWN:
            # if is a pointer
            result += [", ", "dimension"]
            result += ["(", str(settings.array_lower_bound), ":", "*", ")"]
        elif self.variable._shape.dims:
            result += [", ", "dimension"]
            result += get_shape_blocks(
                self.variable._shape.dims, fortran_order=self.variable._shape.fortran_order
            )

        if self.variable.intent is not None:
            result += [", ", "intent", "(", self.variable.intent, ")"]

        if settings.force_value:
            if self.variable._shape is SCALAR and self.variable.intent == "in":
                result += [", ", "value"]

        if self.variable.parameter:
            result += [", ", "parameter"]

        if self.variable.target:
            # why fortran? why?
            if self.variable.pointer:
                result += [", ", "contiguous"]
            else:
                result += [", ", "target"]

        if self.variable.bind_c:
            result += [", ", "bind", "(", "C", ", ", "name=", "'", self.variable.name, "'", ")"]

        result += [" :: ", self.variable.name]

        if self.variable.assign is not None:

            from numeta.syntax.expressions import LiteralNode

            if isinstance(self.variable.assign, (int, float, complex, bool, str)):
                values = LiteralNode(self.variable.assign).get_code_blocks()
            elif isinstance(self.variable.assign, np.ndarray):
                values = []
                for v in self.variable.assign.ravel():
                    values += LiteralNode(v).get_code_blocks()
                    values.append(", ")
                values.pop()
            else:
                raise ValueError("Can only assign scalars or numpy ndarrays")

            if self.variable._shape is UNKNOWN:
                raise ValueError(
                    "Cannot assign to a variable with unknown shape. "
                    "Please specify the shape of the variable."
                )
            else:
                result += [";", " data ", self.variable.name, " / ", *values, " /"]

        return result



import sys

from .nodes import NamedEntity
from .subroutine import Subroutine
from .expressions import Function


class Module(NamedEntity):
    __slots__ = (
        "name",
        "parent",
        "description",
        "hidden",
        "dependencies",
        "derived_types",
        "interfaces",
        "variables",
        "subroutines",
    )

    def __init__(self, name, description=None, hidden=False, parent=None):
        super().__init__(name, parent=parent)
        self.name = name.lower()
        self.description = description
        # hidden define if it is should be a true fortran module or just a container
        self.hidden = hidden

        self.dependencies = {}
        self.derived_types = {}
        self.interfaces = {}
        self.variables = {}
        self.subroutines = {}

    def __getattr__(self, name):
        if name in self.__slots__:  # pragma: no cover
            return self.__getattribute__(name)
        elif name in self.variables:
            return self.variables[name]
        elif name in self.subroutines:
            return self.subroutines[name]
        else:
            raise AttributeError(f"Module {self.name} has no attribute {name}")

    def add_derived_type(self, *derived_types):
        for derived_type in derived_types:
            self.derived_types[derived_type.name] = derived_type
            derived_type.parent = self

    def add_subroutine(self, *subroutines):
        for subroutine in subroutines:
            self.subroutines[subroutine.name] = subroutine
            subroutine.parent = self

    def add_variable(self, *variables):
        for variable in variables:
            self.variables[variable.name] = variable
            variable.parent = self

    def add_interface(self, *subroutines):
        for subroutine in subroutines:
            self.interfaces[subroutine.name] = subroutine

    def get_declaration(self):
        from .statements import ModuleDeclaration

        return ModuleDeclaration(self)

    def print_lines(self, indent=0):
        return self.get_declaration().print_lines(indent=indent)

    def get_code(self):
        return "".join(self.print_lines())

    def get_dependencies(self):
        return self.get_declaration().dependencies


builtins_module = Module(
    "builtins", "The builtins module, to contain built-in functions or subroutines"
)


class ExternalModule(Module):
    """
    **Note**: Only to add support for methods (for external modules).
    When methods will be properly implemented this should be removed
    """

    def __init__(self, name, parent, hidden=False):
        super().__init__(name, hidden=hidden, parent=parent)

    def add_method(self, name, arguments, result_=None, bind_c=False):
        """
        Because currently only subroutines are supported, Modules can only have subroutines.
        But ExternalModule should be able to have functions as well.
        """
        module = self

        if result_ is None:
            # It's a subroutine
            method = Subroutine(name, parent=module, bind_c=bind_c)
            for arg in arguments:
                method.add_variable(arg)
            self.add_subroutine(method)

        else:
            # TODO: Arguments are not used but it could be used to check if the arguments are correct
            def __init__(self, *args):
                from .tools import check_node

                self.name = name
                self.arguments = [check_node(arg) for arg in args]
                self.parent = module

            # to make method pickable
            python_module_name = type(self).__module__
            python_module = sys.modules.get(python_module_name)

            method = type(
                name,
                (Function,),
                {
                    # to make method pickable
                    "__module__": python_module_name,
                    "__init__": __init__,
                    "_ftype": property(lambda self: result_),
                    "_shape": property(lambda self: SCALAR),
                },
            )

            # to make method pickable
            if python_module is not None:
                setattr(python_module, name, method)

            self.subroutines[name] = method



import numpy as np
import numeta as nm
from numeta.syntax.settings import settings


def test_slice_exclusive_bounds():

    @nm.jit
    def fill(a):
        a[1:5] = 1

    arr = np.zeros(6, dtype=np.int64)
    fill(arr)

    expected = np.zeros(6, dtype=np.int64)
    expected[1:5] = 1
    np.testing.assert_array_equal(arr, expected)


def test_slice_inclusive_bounds():

    settings.unset_c_like_bounds()

    @nm.jit
    def fill(a):
        a[:4] = 1

    arr = np.zeros(6, dtype=np.int64)
    fill(arr)

    settings.set_c_like_bounds()

    expected = np.zeros(6, dtype=np.int64)
    expected[:5] = 1
    np.testing.assert_array_equal(arr, expected)


test_slice_inclusive_bounds()



import numpy as np
from numeta.builder_helper import BuilderHelper
from numeta.datatype import DataType, float64, FortranType, get_datatype
from numeta.array_shape import ArrayShape


def empty(shape, dtype: DataType | FortranType | np.generic = float64, order="C", name=None):
    if order not in ["C", "F"]:
        raise ValueError(f"Invalid order: {order}, must be 'C' or 'F'")

    fortran_order = order == "F"
    if not isinstance(shape, ArrayShape):
        if not isinstance(shape, (tuple, list)):
            shape = (shape,)
        shape = ArrayShape(shape, fortran_order=fortran_order)

    if isinstance(dtype, FortranType):
        ftype = dtype
    else:
        ftype = get_datatype(dtype).get_fortran()

    allocate = shape.has_comptime_undefined_dims()
    array = BuilderHelper.generate_local_variables(
        "fc_a",
        name=name,
        ftype=ftype,
        shape=shape,
        allocate=allocate,
    )

    return array



import numpy as np
from numeta.datatype import DataType, float64
from numeta.syntax import FortranType
from .empty import empty


def zeros(shape, dtype: DataType | FortranType | np.generic = float64, order="C", name=None):
    array = empty(shape, dtype=dtype, order=order, name=name)
    array[:] = 0.0
    return array



import numeta as nm
import numpy as np
import tempfile
import shutil


def test_dump_load():
    n = 100

    a = np.random.rand(n, n)

    dumpdir = tempfile.mkdtemp()
    tmpdir = tempfile.mkdtemp()

    @nm.jit(directory=tmpdir)
    def fill(value: nm.comptime, a):
        a[:] = value

    fill(1.0, a)

    fill.dump(dumpdir)

    shutil.rmtree(tmpdir)

    @nm.jit
    def fill(value: nm.comptime, a):
        raise Warning("This should not be called")

    fill.load(dumpdir)

    fill(1.0, a)

    np.testing.assert_allclose(a, np.ones((n, n)), rtol=10e2 * np.finfo(np.float64).eps)



from numeta.syntax.tools import check_node
from numeta.syntax.scope import Scope
from numeta.syntax.settings import settings
from .tools import print_block, get_shape_blocks
from .statement import Statement, StatementWithScope


class Comment(Statement):
    def __init__(self, comment, add_to_scope=False):
        super().__init__(add_to_scope=add_to_scope)
        self.comment = comment
        if isinstance(comment, str):
            self.comment = [comment]

    @property
    def children(self):
        return []

    def print_lines(self, indent=0):
        """Print the statement, formatted with the given indent level."""
        return [print_block(self.comment, indent=indent, prefix="! ")]


class Use(Statement):
    def __init__(self, module, only=None, add_to_scope=True):
        super().__init__(add_to_scope=add_to_scope)
        self.module = module
        self.only = only

    @property
    def children(self):
        return []

    def get_code_blocks(self):
        result = ["use", " ", self.module.name]
        if self.only is not None:
            result += [", ", "only", ": ", self.only.name]
        return result


class Implicit(Statement):
    def __init__(self, implicit_type="none", add_to_scope=True):
        super().__init__(add_to_scope=add_to_scope)
        self.implicit_type = implicit_type

    @property
    def children(self):
        return []

    def get_code_blocks(self):
        return ["implicit", " ", self.implicit_type]


class Assignment(Statement):
    def __init__(self, assignment_target, value, add_to_scope=True):
        super().__init__(add_to_scope=add_to_scope)
        self.target = check_node(assignment_target)
        self.value = check_node(value)

    @property
    def children(self):
        return [self.target, self.value]

    def get_code_blocks(self):
        result = self.target.get_code_blocks()
        result.append("=")
        result += self.value.get_code_blocks()
        return result


class SimpleStatement(Statement):
    token = ""

    def __init__(self):
        Scope.add_to_current_scope(self)

    @property
    def children(self):
        return []

    def get_code_blocks(self):
        return [self.__class__.token]


class Cycle(SimpleStatement):
    token = "cycle"


class Exit(SimpleStatement):
    token = "exit"


class Stop(SimpleStatement):
    token = "stop"


class Return(SimpleStatement):
    token = "return"


class Print(Statement):
    def __init__(self, *children, add_to_scope=True):
        super().__init__(add_to_scope=add_to_scope)
        self.to_print = [check_node(child) for child in children]

    @property
    def children(self):
        return self.to_print

    def get_code_blocks(self):
        result = ["print *, "]
        for child in self.children:
            if isinstance(child, str):
                protected_child = child.replace('"', '""')
                result.append(f'"{protected_child}"')
            else:
                result += child.get_code_blocks()
            result.append(", ")
        result.pop()
        return result


class Allocate(Statement):
    def __init__(self, target, *shape):
        super().__init__()
        self.target = check_node(target)
        self.shape = [check_node(child) for child in shape]

    @property
    def children(self):
        return [self.target] + self.shape

    def get_code_blocks(self):
        result = ["allocate", "("]

        result += self.target.get_code_blocks()

        dims = []
        for argument in self.shape:
            if (lbound := settings.array_lower_bound) != 1:
                dims.append([str(lbound), ":", *(argument + (lbound - 1)).get_code_blocks()])
            else:
                dims.append([*(argument).get_code_blocks()])

        if not self.target._shape.fortran_order:
            dims = dims[::-1]

        result.append("(")
        result += dims[0]
        for dim in dims[1:]:
            result += [",", " "]
            result += dim
        result.append(")")

        result.append(")")

        return result

        # contains some lists
        # new_result = []
        # for element in result:
        #    if type(element) is list:
        #        new_result += element
        #    else:
        #        new_result.append(element)

        # return new_result


class Deallocate(Statement):
    def __init__(self, array, add_to_scope=True):
        super().__init__(add_to_scope=add_to_scope)
        self.array = check_node(array)

    @property
    def children(self):
        return [self.array]

    def get_code_blocks(self):
        result = ["deallocate", "(", *self.array.get_code_blocks(), ")"]
        return result


class Do(StatementWithScope):
    def __init__(self, iterator, start, end, /, step=None, *, add_to_scope=True, enter_scope=True):
        super().__init__(add_to_scope=add_to_scope, enter_scope=enter_scope)
        self.iterator = check_node(iterator)
        self.start = check_node(start)
        self.end = check_node(end)
        self.step = None if step is None else check_node(step)

    @property
    def children(self):
        return [self.iterator, self.start, self.end] + (
            [self.step] if self.step is not None else []
        )

    def get_start_code_blocks(self):
        result = ["do", " "]
        result += self.iterator.get_code_blocks()
        result += [" ", "=", " "]
        result += self.start.get_code_blocks()
        result.append(", ")
        result += self.end.get_code_blocks()
        if len(self.children) == 4:
            result.append(", ")
            result += self.step.get_code_blocks()

        return result

    def get_end_code_blocks(self):
        return ["end", " ", "do"]


class DoWhile(StatementWithScope):
    def __init__(self, condition, /, *, add_to_scope=True, enter_scope=True):
        super().__init__(add_to_scope=add_to_scope, enter_scope=enter_scope)
        self.condition = check_node(condition)

    @property
    def children(self):
        return [self.condition]

    def get_start_code_blocks(self):
        return ["do while", " ", "(", *self.condition.get_code_blocks(), ")"]

    def get_end_code_blocks(self):
        return ["end", " ", "do"]


class If(StatementWithScope):
    def __init__(self, condition, /, *, add_to_scope=True, enter_scope=True):
        super().__init__(add_to_scope=add_to_scope, enter_scope=enter_scope)
        self.condition = check_node(condition)
        self.orelse = []

    @property
    def children(self):
        return [self.condition]

    def get_statements(self):
        return self.scope.get_statements() + self.orelse

    def print_lines(self, indent=0):
        """Print the entire scoped statement, including all nested statements."""
        result = [print_block(self.get_start_code_blocks(), indent=indent)]
        for statement in self.scope.get_statements():
            result.extend(statement.print_lines(indent=indent + 1))
        for statement in self.orelse:
            result.extend(statement.print_lines(indent=indent))
        result.append(print_block(self.get_end_code_blocks(), indent=indent))
        return result

    def get_start_code_blocks(self):
        return ["if", "(", *self.condition.get_code_blocks(), ")", "then"]

    def get_end_code_blocks(self):
        return ["end", " ", "if"]

    def get_with_updated_variables(self, variables_couples):
        new_children = [
            child.get_with_updated_variables(variables_couples) for child in self.children
        ]
        result = type(self)(*new_children, add_to_scope=False, enter_scope=False)
        result.scope = self.scope.get_with_updated_variables(variables_couples)
        result.orelse = [stmt.get_with_updated_variables(variables_couples) for stmt in self.orelse]
        return result


class ElseIf(StatementWithScope):
    def __init__(self, condition, /, *, add_to_scope=True, enter_scope=True):
        if add_to_scope:
            if not isinstance(Scope.current_scope.body[-1], If):
                raise Exception(
                    "Something went wrong with this else if. The last statement is not an if statement."
                )
            Scope.current_scope.body[-1].orelse.append(self)
        self.scope = Scope()
        if enter_scope:
            self.scope.enter()
        self.condition = check_node(condition)

    @property
    def children(self):
        return [self.condition]

    def get_start_code_blocks(self):
        return ["elseif", "(", *self.condition.get_code_blocks(), ")", "then"]

    def get_end_code_blocks(self):
        return []


class Else(StatementWithScope):
    def __init__(self, /, *, add_to_scope=True, enter_scope=True):
        if add_to_scope:
            if not isinstance(Scope.current_scope.body[-1], If):
                raise Exception(
                    "Something went wrong with this else if. The last statement is not an if statement."
                )
            Scope.current_scope.body[-1].orelse.append(self)
        self.scope = Scope()
        if enter_scope:
            self.scope.enter()

    @property
    def children(self):
        return []

    def get_start_code_blocks(self):
        return ["else"]

    def get_end_code_blocks(self):
        return []


class SelectCase(StatementWithScope):
    def __init__(self, value, add_to_scope=True):
        super().__init__(add_to_scope=add_to_scope)
        self.value = check_node(value)

    @property
    def children(self):
        return [self.value]

    def get_start_code_blocks(self):
        return ["select", " ", "case", " ", "(", *self.value.get_code_blocks(), ")"]

    def get_end_code_blocks(self):
        return ["end", " ", "select"]


class Case(StatementWithScope):
    def __init__(self, value, add_to_scope=True):
        super().__init__(add_to_scope=add_to_scope)
        self.value = check_node(value)

    @property
    def children(self):
        return [self.value]

    def get_start_code_blocks(self):
        result = ["case", " ", "("]
        result += self.value.get_code_blocks()
        result.append(")")
        return result

    def get_end_code_blocks(self):
        return []


class Contains(Statement):
    def __init__(self, add_to_scope=True):
        super().__init__(add_to_scope=add_to_scope)

    @property
    def children(self):
        return []

    def get_code_blocks(self):
        return ["contains"]


class Interface(StatementWithScope):
    def __init__(self, methods):
        super().__init__(add_to_scope=False)
        self.methods = methods

    @property
    def children(self):
        return []

    def get_statements(self):
        result = []
        for method in self.methods:
            result.append(method.get_interface_declaration())
        return result

    def get_start_code_blocks(self):
        return ["interface"]

    def get_end_code_blocks(self):
        return ["end", " ", "interface"]


class PointerAssignment(Statement):
    def __init__(self, pointer, pointer_shape, target, target_shape=None, add_to_scope=True):
        super().__init__(add_to_scope=add_to_scope)
        self.pointer = check_node(pointer)
        self.pointer_shape = []
        # should specify bounds for the pointer
        for dim in pointer_shape.dims:
            if not isinstance(dim, slice):
                self.pointer_shape.append(slice(None, dim))
            else:
                self.pointer_shape.append(dim)
        if not pointer_shape.fortran_order:
            self.pointer_shape.reverse()
        self.pointer_shape = tuple(self.pointer_shape)
        self.target = check_node(target)
        self.target_shape = target_shape

        from numeta.syntax.variable import Variable
        from numeta.syntax.expressions import GetItem

        if isinstance(self.target, Variable):
            self.target.target = True
        elif isinstance(self.target, GetItem):
            self.target.variable.target = True
        else:
            raise Exception("The target of a pointer must be a variable or GetItem.")

        if not isinstance(self.pointer, Variable):
            raise Exception("The pointer must be a variable.")
        self.pointer.pointer = True

    @property
    def children(self):
        return [self.target, self.pointer]

    def get_with_updated_variables(self, variables_couples):
        def update_variables(element):
            if isinstance(element, tuple):
                return tuple(update_variables(e) for e in element)
            if isinstance(element, slice):
                return slice(
                    update_variables(element.start),
                    update_variables(element.stop),
                    update_variables(element.step),
                )
            from numeta.array_shape import ArrayShape

            if isinstance(element, ArrayShape):
                return ArrayShape(
                    tuple(update_variables(dim) for dim in element.dims),
                    fortran_order=element.fortran_order,
                )
            from numeta.syntax.nodes.base_node import Node

            if isinstance(element, Node):
                return element.get_with_updated_variables(variables_couples)
            return element

        new_pointer = self.pointer.get_with_updated_variables(variables_couples)
        new_target = self.target.get_with_updated_variables(variables_couples)
        from numeta.array_shape import ArrayShape

        new_pointer_shape = ArrayShape(update_variables(self.pointer_shape), fortran_order=True)
        new_target_shape = (
            update_variables(self.target_shape) if self.target_shape is not None else None
        )
        return type(self)(
            new_pointer,
            new_pointer_shape,
            new_target,
            new_target_shape,
            add_to_scope=False,
        )

    def get_code_blocks(self):
        return [
            *self.pointer.get_code_blocks(),
            *get_shape_blocks(self.pointer_shape),
            "=>",
            *self.target.get_code_blocks(),
        ]



from numeta.syntax import ExternalModule
from numeta.syntax.settings import settings as syntax_settings


class OmpModule(ExternalModule):
    def __init__(self):
        super().__init__("omp_lib", None)
        self.add_method("omp_get_thread_num", arguments=[], result_=syntax_settings.DEFAULT_INTEGER)
        self.add_method(
            "omp_get_max_threads", arguments=[], result_=syntax_settings.DEFAULT_INTEGER
        )

    def do(self, *args, **kwargs):
        from numeta.syntax import Do, Comment
        from numeta.syntax.statements.tools import print_block

        class OmpComment(Comment):
            """
            Hack to handle OpenMP statements
            """

            def print_lines(self, indent=0):
                return [print_block(self.comment, indent=indent, prefix="!$omp ")]

        class OmpDo(Do):
            def __init__(
                self,
                *args,
                default="private",
                schedule="dynamic",
                private=None,
                shared=None,
                **kwargs,
            ):
                omp_code = ["parallel do", " "]
                omp_code += [f"default({default})", " "]
                omp_code += [f"schedule({schedule})", " "]
                if private is not None:
                    # TODO: Should we add automatically variables declared inside the loop?
                    omp_code += ["private("]
                    for var in private:
                        omp_code += var.get_code_blocks()
                        omp_code += [", "]
                    omp_code[-1] = ")"
                if shared is not None:
                    # TODO: Should we automatically array shapes if explicitly declared as variables?
                    omp_code += ["shared("]
                    for var in shared:
                        omp_code += var.get_code_blocks()
                        omp_code += [", "]
                    omp_code[-1] = ")"
                OmpComment(omp_code, add_to_scope=True)
                super().__init__(*args, **kwargs)

            def __exit__(self, exc_type, exc_value, traceback):
                super().__exit__(exc_type, exc_value, traceback)
                OmpComment("end parallel do", add_to_scope=True)

        return OmpDo(*args, **kwargs)

    def Do(self, *args, **kwargs):
        return self.do(*args, **kwargs)

    def DO(self, *args, **kwargs):
        return self.do(*args, **kwargs)

    def atomic_update_op(self, variable, to_assign, op):
        from numeta.syntax.statements import Assignment, Comment
        from numeta.syntax.expressions.binary_operation_node import (
            BinaryOperationNodeNoPar,
        )

        Comment(f"$omp atomic update", add_to_scope=True)
        Assignment(
            variable,
            BinaryOperationNodeNoPar(variable, op, to_assign),
            add_to_scope=True,
        )

    def atomic_update_add(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "+")

    def AtomicUpdateAdd(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "+")

    def ATOMIC_UPDATE_ADD(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "+")

    def atomic_update_sub(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "-")

    def AtomicUpdateSub(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "-")

    def ATOMIC_UPDATE_SUB(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "-")

    def atomic_update_mul(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "*")

    def AtomicUpdateMul(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "*")

    def ATOMIC_UPDATE_MUL(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "*")

    def atomic_update_div(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "/")

    def AtomicUpdateDiv(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "/")

    def ATOMIC_UPDATE_DIV(self, variable, to_assign):
        self.atomic_update_op(variable, to_assign, "/")


omp = OmpModule()



from .expression_node import ExpressionNode
from numeta.syntax.tools import check_node
from numeta.array_shape import ArrayShape, UNKNOWN, SCALAR


class BinaryOperationNode(ExpressionNode):
    __slots__ = ["op", "left", "right"]

    def __init__(self, left, op, right):
        self.op = op
        # self.left = left
        # self.right = right
        self.left = check_node(left)
        self.right = check_node(right)

    @property
    def _ftype(self):
        """Return the Fortran type of the expression."""
        # This is a simplification. In reality, the type of the result
        # depends on the types of the operands and the operation.
        # For example, dividing two integers should result in a real.
        # For now, we'll just return the type of the left operand.
        return self.left._ftype

    @property
    def _shape(self):
        """Return the shape of the expression if any."""
        # This is a simplification. It doesn't handle broadcasting correctly.
        # For now, we'll just return the shape of the left operand.
        if self.left._shape is SCALAR:
            return self.right._shape
        elif self.right._shape is SCALAR:
            return self.left._shape
        else:
            return self.left._shape

    def get_code_blocks(self):
        result = ["("]
        # if hasattr(self.left, "get_code_blocks"):
        #    result += self.left.get_code_blocks()
        # else:
        #    from .literal_node import LiteralNode
        #    result += LiteralNode(self.left).get_code_blocks()
        result += self.left.get_code_blocks()
        result.append(self.op)
        # if hasattr(self.right, "get_code_blocks"):
        #    result += self.right.get_code_blocks()
        # else:
        #    from .literal_node import LiteralNode
        #    result += LiteralNode(self.right).get_code_blocks()
        result += self.right.get_code_blocks()
        result.append(")")
        return result

    def get_with_updated_variables(self, variables_couples):
        return BinaryOperationNode(
            self.left.get_with_updated_variables(variables_couples),
            self.op,
            self.right.get_with_updated_variables(variables_couples),
        )

    def extract_entities(self):
        # if hasattr(self.left, "extract_entities"):
        #    yield from self.left.extract_entities()
        # if hasattr(self.right, "extract_entities"):
        #    yield from self.right.extract_entities()
        yield from self.left.extract_entities()
        yield from self.right.extract_entities()


class BinaryOperationNodeNoPar(BinaryOperationNode):
    def get_code_blocks(self):
        return [*self.left.get_code_blocks(), self.op, *self.right.get_code_blocks()]


class EqBinaryNode(BinaryOperationNode):
    __slots__ = ["op", "left", "right"]

    def __init__(self, left, right):
        # faster than calling super().__init__(left, '.eq.', right)
        self.op = ".eq."
        self.left = check_node(left)
        self.right = check_node(right)

    def __bool__(self):
        try:
            return self.left.name == self.right.name
        except AttributeError:
            raise Warning(
                f"Do not use '==' operator for non-NamedEntity: {type(self.left)} and {type(self.right)}"
            )
        # TODO: Too slow

    # from numeta.syntax.named_entity import NamedEntity
    ##Always raise Warning except for the case when we evaluating two variables
    # if isinstance(self.left, NamedEntity) and isinstance(self.right, NamedEntity):
    #    return self.left.name == self.right.name
    # raise Warning(f"Do not use '==' operator for non-NamedEntity: {type(self.left)} and {type(self.right)}")


class NeBinaryNode(BinaryOperationNode):
    def __init__(self, left, right):
        self.op = ".ne."
        # self.left = left
        # self.right = right
        self.left = check_node(left)
        self.right = check_node(right)

    def __bool__(self):
        try:
            return self.left.name != self.right.name
        except AttributeError:
            raise Warning(
                f"Do not use '!=' operator for non-NamedEntity: {type(self.left)} and {type(self.right)}"
            )

        # TODO: Too slow

    # from numeta.syntax.named_entity import NamedEntity

    ## Always raise Warning except for the case when we evaluating two variables
    # if isinstance(self.left, NamedEntity) and isinstance(self.right, NamedEntity):
    #    return self.left.name != self.right.name

    # raise Warning(f"Do not use '!=' operator for non-NamedEntity: {type(self.left)} and {type(self.right)}")



from .nodes import NamedEntity


class DerivedType(NamedEntity):
    """
    A derived type. Actually used to define structs.

    Parameters
    ----------
    name : str
        The name of the derived type.
    fields : list of tuples
        The fields of the derived type, each tuple containing the name, datatype, and dimension.
    """

    def __init__(self, name, fields):
        super().__init__(name)
        self.fields = fields
        for name, _, shape in self.fields:
            if shape.has_comptime_undefined_dims():
                raise ValueError(
                    f"Derived type '{name}' cannot have compile-time undefined dimensions."
                )
        self.parent = None

    def get_declaration(self):
        from .statements import DerivedTypeDeclaration

        return DerivedTypeDeclaration(self)



import ast
import inspect
from numeta.syntax import IF, ELSE, ELSEIF, Scope


class CondHelper:
    curr = None

    @classmethod
    def get_instance(cls):
        if cls.curr is None:
            raise SyntaxError("No if created yet.")
        return cls.curr

    @classmethod
    def set_instance(cls, instance):
        cls.curr = instance

    def __init__(self):
        self.number_if = 0
        self.if_stack = []
        self.source_cache = {}

    def add_if(self, lineno):
        self.if_stack.append(lineno)
        self.number_if += 1

    def pop_if(self):
        return self.if_stack.pop()

    def get_source_cache(self, frame):
        if frame.f_code.co_filename.startswith("<fc_if_"):
            if_id = int(frame.f_code.co_filename.split("_")[-1].split(">")[0])
            source_lines, tree, starting_line_number = self.source_cache[if_id]
            endif_line_number = frame.f_lineno
            if_line_number = self.if_stack.pop()
        else:
            if_id = self.number_if
            self.number_if += 1
            source_lines, starting_line_number = inspect.getsourcelines(frame.f_code)
            # HACK: add a try except block to take care of the indentation
            source_lines = (
                ["try:\n"]
                + source_lines
                + ["except:\n    raise Warning('impossible to parse the code')"]
            )
            tree = ast.parse("".join(source_lines))
            self.source_cache[if_id] = (source_lines, tree, starting_line_number + 1)
            endif_line_number = frame.f_lineno - starting_line_number + 2
            if_line_number = self.if_stack.pop() - starting_line_number + 2

        return source_lines, tree, if_id, if_line_number, endif_line_number

    def analyze_frame_and_get_if(self, frame):
        """
        Analyze the frame, sets the internal variables
        and returns:
            - True if it is an if statement
            - False if it is an elif statement
        """
        if not frame.f_code.co_filename.startswith("<fc_if_"):
            # it is surely an if statement bc it is not called from the ast exec
            # All the elif statements are called from the ast exec
            self.if_stack.append(frame.f_lineno)
            return True

        else:
            # called from the ast exec so it can be an elif statement
            if_id = int(frame.f_code.co_filename.split("_")[-1].split(">")[0])
            source_lines, _, _ = self.source_cache[if_id]

            if source_lines[frame.f_lineno - 1].strip().startswith("elif"):
                return False
            else:
                self.if_stack.append(frame.f_lineno)
                return True

    def finalize(self):
        if len(self.if_stack) > 0:
            raise SyntaxError("Unmatched if statement")


_helper = CondHelper()
CondHelper.set_instance(_helper)


def cond(cond):
    frame = inspect.currentframe().f_back
    helper = CondHelper.get_instance()
    is_if = helper.analyze_frame_and_get_if(frame)

    if is_if:
        IF(cond)
    else:
        ELSEIF(cond)

    return True


def endif():
    frame = inspect.currentframe().f_back

    helper = CondHelper.get_instance()

    (
        source_lines,
        tree,
        if_id,
        if_line_number,
        endif_line_number,
    ) = helper.get_source_cache(frame)

    previous_if = None
    for node in ast.walk(tree):
        if isinstance(node, ast.If) and node.lineno == if_line_number:
            previous_if = node
            break

    from collections import deque

    def get_previous_if(node):
        # BFS to find the previous if statement in the tree wrt node

        queue = deque([node])
        while queue:
            current_node = queue.popleft()
            for child in ast.iter_child_nodes(current_node):
                if isinstance(child, ast.Expr) and child.lineno == endif_line_number:
                    return queue[-1]
                queue.append(child)

    if get_previous_if(tree.body[0]).lineno != previous_if.lineno:
        msg = "Unmatched if statement for an endif()\n"
        msg += "Check this if statement\n"
        msg += ast.unparse(previous_if)
        raise SyntaxError(msg)

    write_elif(previous_if, source_lines, frame, if_id)

    # Important: close the scope of the if statement
    Scope.end()


def write_elif(current_if, source_lines, frame, if_id):
    while hasattr(current_if, "orelse") and len(current_if.orelse) != 0:
        else_body = current_if.orelse

        if len(else_body) != 1 or not isinstance(else_body[0], ast.If):
            # it is surely an else statement
            treat_else(else_body, frame, if_id)

            break
        else:
            """
            If the length of the orelse is 1 and it contains an if statement,
            it is probably an elif statement.
            But it could also be an else statement like this:

            else:
                if True:
                    print('hey')

            In this case, the orelse will contain an if statement.
            So we have to check the source code to see if it is an elif or else statement
            """
            current_if = else_body[0]

            if source_lines[current_if.lineno - 1].strip().startswith("elif"):
                # It is an elif statement

                # close the previous scope
                Scope.end()

                # execute the condition and the body
                compiled_cond = compile(
                    ast.Module([current_if]), filename=f"<fc_if_{if_id}>", mode="exec"
                )
                exec(compiled_cond, frame.f_globals, frame.f_locals)

                # to update the locals
                import ctypes

                ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_int(0))

            else:
                treat_else(else_body, frame, if_id)
                # it is an else statement
                break


def treat_else(else_body, frame, if_id):
    # close the previous scope
    Scope.end()

    ELSE()

    compiled_if_body = compile(ast.Module(else_body), filename=f"<fc_if_{if_id}>", mode="exec")
    exec(compiled_if_body, frame.f_globals, frame.f_locals)

    # to update the locals
    import ctypes

    ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_int(0))



import numpy as np

from .expression_node import ExpressionNode
from numeta.syntax.settings import settings
from numeta.array_shape import SCALAR


class LiteralNode(ExpressionNode):
    __slots__ = ["value", "__ftype"]

    def __init__(self, value):
        self.value = value
        if isinstance(value, (bool, np.bool_)):
            # IMPORTANT before int because bool is a subclass of int
            self.__ftype = settings.DEFAULT_LOGICAL
        elif isinstance(value, (int, np.int32, np.int64)):
            self.__ftype = settings.DEFAULT_INTEGER
        elif isinstance(value, (float, np.float64, np.float32)):
            self.__ftype = settings.DEFAULT_REAL
        elif isinstance(value, (complex, np.complex64, np.complex128)):
            self.__ftype = settings.DEFAULT_COMPLEX
        elif isinstance(value, str):
            self.__ftype = settings.DEFAULT_CHARACTER
        else:
            raise ValueError(
                f"Type {value.__class__.__name__} is unsupported for LiteralNode,\n value: {value}"
            )

    @property
    def _ftype(self):
        return self.__ftype

    @property
    def _shape(self):
        return SCALAR

    def extract_entities(self):
        yield from self._ftype.extract_entities()

    def get_with_updated_variables(self, variables_couples):
        return self

    def get_code_blocks(self):
        kind = self._ftype.get_kind_str()

        if self._ftype.type == "type":
            return [f"{self.value}"]
        elif self._ftype.type == "integer":
            return [f"{int(self.value)}_{kind}"]
        elif self._ftype.type == "real":
            return [f"{float(self.value)}_{kind}"]
        elif self._ftype.type == "complex":
            return [
                "(",
                f"{self.value.real}_{kind}",
                "," f"{self.value.imag}_{kind}",
                ")",
            ]
        elif self._ftype.type == "logical":
            if self.value is True:
                return [f".true._{kind}"]
            else:
                return [f".false._{kind}"]
        elif self._ftype.type == "character":
            return [f'"{self.value}"']
        else:
            raise ValueError(f"Unknown type: {self._ftype.type}")



from .statement import StatementWithScope
from .variable_declaration import VariableDeclaration
from numeta.syntax.variable import Variable
from numeta.syntax.settings import settings


class DerivedTypeDeclaration(StatementWithScope):
    def __init__(self, derived_type):
        super().__init__(enter_scope=False, add_to_scope=False)
        self.derived_type = derived_type

    @property
    def children(self):
        return []

    def get_statements(self):
        for name, fortran_type, shape in self.derived_type.fields:
            yield VariableDeclaration(Variable(name, ftype=fortran_type, shape=shape))

    def get_start_code_blocks(self):
        if settings.derived_type_bind_c:
            return ["type", ", ", "bind(C)", " ", "::", " ", self.derived_type.name]
        return ["type", " ", "::", " ", self.derived_type.name]

    def get_end_code_blocks(self):
        return ["end", " ", "type", " ", self.derived_type.name]



import numpy as np
from numeta.builder_helper import BuilderHelper
from numeta.datatype import DataType, FortranType, get_datatype
from numeta.array_shape import ArrayShape


def constant(value, dtype: DataType | FortranType | np.generic | None = None, order="C", name=None):
    if order not in ["C", "F"]:
        raise ValueError(f"Invalid order: {order}, must be 'C' or 'F'")

    builder = BuilderHelper.get_current_builder()

    # Determine value shape and numpy representation
    if isinstance(value, np.ndarray):
        np_value = value
    elif isinstance(value, (list, tuple)):
        np_value = np.array(value)
    else:
        np_value = None

    if np_value is not None:
        shape = np_value.shape
        fortran_order = (
            np_value.flags.f_contiguous if isinstance(value, np.ndarray) else order == "F"
        )
        # Determine datatype if not provided
        if dtype is None:
            dtype = DataType.from_np_dtype(np_value.dtype.type)
        assign_value = np_value
    else:
        shape = None
        fortran_order = True
        if dtype is None:
            np_dtype = np.array(value).dtype.type
            dtype = DataType.from_np_dtype(np_dtype)
        assign_value = value

    # Convert dtype to FortranType
    if isinstance(dtype, FortranType):
        ftype = dtype
    else:
        ftype = get_datatype(dtype).get_fortran()

    if name is None:
        name = "fc_c"

    if shape is None or shape == ():
        if isinstance(assign_value, (int, float, complex)):
            assign_final = str(assign_value)
        else:
            assign_final = assign_value

        return builder.generate_local_variables(
            name,
            ftype=ftype,
            # TODO
            # parameter=True, # parameter is not supported yet, so not really constant.
            # parameter=True,
            assign=assign_final,
        )

    array_shape = ArrayShape(shape, fortran_order=fortran_order)
    return builder.generate_local_variables(
        name,
        ftype=ftype,
        shape=array_shape,
        # TODO
        # parameter=True, # parameter is not supported yet, so not really constant.
        # parameter=True,
        assign=assign_value,
    )



import numeta as nm
import numpy as np
import pytest


def test_comptime_type_hint():
    n = 100

    dtype = np.float64

    @nm.jit
    def mul(dtype: nm.comptime, a, b, c):
        for i in nm.range(a.shape[0]):
            for k in nm.range(b.shape[0]):
                c[i, :] += a[i, k] * b[k, :]

    a = np.random.rand(n, n).astype(dtype)
    b = np.random.rand(n, n).astype(dtype)
    c = np.zeros((n, n), dtype=dtype)

    mul(np.float64, a, b, c)

    np.testing.assert_allclose(c, np.dot(a, b), rtol=10e2 * np.finfo(dtype).eps)

    dtype = np.complex128

    a = np.random.rand(n, n).astype(dtype)
    b = np.random.rand(n, n).astype(dtype)
    c = np.zeros((n, n), dtype=dtype)
    mul(np.complex128, a, b, c)

    np.testing.assert_allclose(c, np.dot(a, b), rtol=10e2 * np.finfo(dtype).eps)



import numeta as nm
import numpy as np


def test_custom_namer():
    @nm.jit(namer=lambda length, *_: f"spec_{length}")
    def fill(length: nm.comptime, a):
        for i in range(length):
            a[i] = i

    arr = np.zeros(5, dtype=np.int64)
    fill(3, arr)

    assert fill.get_symbolic_functions()[0].name == "spec_3"
    np.testing.assert_allclose(arr[:3], np.arange(3))



from dataclasses import dataclass
from typing import Optional, Tuple


@dataclass(frozen=True)
class ArrayShape:
    # None      ⇒ unknown shape
    # ()        ⇒ scalar
    # (3,4)     ⇒ fixed 2-D So it is know at compile time
    # (None, 4) ⇒ 2-D with an undefined dimension at compile time
    _dims: Optional[Tuple]
    fortran_order: bool = False

    def __repr__(self) -> str:
        if self._dims is None:
            return "ArrayShape<unknown>"
        elif len(self._dims) == 0:
            return "ArrayShape<scalar>"
        else:
            inner = ", ".join(map(str, self.dims))
            return f"ArrayShape[{inner}]"

    @property
    def dims(self) -> Tuple:
        if self._dims is None:
            raise ValueError("Cannot get comptime undefined dimensions for unknown shape.")
        return self._dims

    @property
    def rank(self) -> int:
        """
        Returns the rank of the array shape.
        """
        return len(self.dims)

    @property
    def comptime_undefined_dims(self):
        """
        Returns the indices of the dimensions that are undefined at compile time.
        """
        return [i for i, dim in enumerate(self.dims) if isinstance(dim, int)]

    def has_comptime_undefined_dims(self):
        """
        Checks if the argument has undefined dimensions at compile time.
        """
        if self._dims is None:
            # The dimensions are unknown
            return False
        for dim in self.dims:
            if not isinstance(dim, int):
                return True
        return False


# sentinels
UNKNOWN = ArrayShape(None)
SCALAR = ArrayShape(())



from .expression_node import ExpressionNode


class GetAttr(ExpressionNode):
    def __init__(self, variable, attr):
        self.variable = variable
        self.attr = attr

    @property
    def _ftype(self):
        derived_type = self.variable._ftype.kind
        for name, fortran_type, _ in derived_type.fields:
            if name == self.attr:
                return fortran_type
        raise ValueError(f"Attribute '{self.attr}' not found in derived type '{derived_type.name}'")

    @property
    def _shape(self):
        derived_type = self.variable._ftype.kind
        for name, _, shape in derived_type.fields:
            if name == self.attr:
                return shape
        raise ValueError(f"Attribute '{self.attr}' not found in derived type '{derived_type.name}'")

    def get_code_blocks(self):
        return [*self.variable.get_code_blocks(), "%", self.attr]

    def extract_entities(self):
        yield from self.variable.extract_entities()

    def get_with_updated_variables(self, variables_couples):
        return GetAttr(self.variable.get_with_updated_variables(variables_couples), self.attr)

    def __setitem__(self, key, value):
        """Does nothing, but allows to use variable[key] = value"""
        from numeta.syntax.statements import Assignment

        if isinstance(key, slice) and key.start is None and key.stop is None and key.step is None:
            # if the variable is assigned to itself, do nothing, needed for the += and -= operators
            if self is value:
                return
            Assignment(self, value)
        else:
            Assignment(self[key], value)



import numeta as nm
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_reshape(dtype):
    n = 100
    m = 20

    @nm.jit
    def set_zero_first_col(n, m, a):
        a_p = nm.reshape(a, (n, m))

        a_p[:, 0] = 0

    a = np.random.rand(n, m).astype(dtype)
    b = a.copy()

    set_zero_first_col(n, m, a)
    b[:, 0] = 0

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(a, b, atol=0)
    else:
        np.testing.assert_allclose(a, b, rtol=10e2 * np.finfo(dtype).eps)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_reshape_fortran(dtype):
    n = 100
    m = 20

    @nm.jit
    def set_zero_first_col(n, m, a):
        a_p = nm.reshape(a, (n, m), order="F")

        a_p[:, 0] = 0

    a = np.asfortranarray(np.random.rand(n, m).astype(dtype))
    b = a.copy()

    set_zero_first_col(n, m, a)
    b[:, 0] = 0

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(a, b, atol=0)
    else:
        np.testing.assert_allclose(a, b, rtol=10e2 * np.finfo(dtype).eps)



from abc import ABC, abstractmethod


class Node(ABC):
    @abstractmethod
    def get_code_blocks(self):
        """Return the code blocks representing the node."""
        raise NotImplementedError(
            f"Subclass '{self.__class__.__name__}' must implement 'get_code_blocks'."
        )

    def __str__(self):
        return "".join(self.get_code_blocks())

    @abstractmethod
    def extract_entities(self):
        """Extract the nested entities of the node."""



from .datatype import (
    DataType,
    StructType,
    int32,
    int64,
    float32,
    float64,
    complex64,
    complex128,
    bool8,
    char,
    size_t,
    get_datatype,
)
from .types_hint import comptime

integer4 = int32
integer8 = int64
i4 = int32
i8 = int64

real4 = float32
real8 = float64
f4 = float32
f8 = float64
r4 = float32
r8 = float64

complex8 = complex64
complex16 = complex128
c8 = complex64
c16 = complex128

logical1 = bool8
b1 = bool8

from .external_modules import iso_c, omp

from .jit import jit, jitted_functions, clear_jitted_functions
from .wrappers import *
from .syntax import *
from .settings import settings



from numeta.syntax import Variable, FortranType, ExternalModule
from numeta.external_library import ExternalLibrary
from numeta.datatype import DataType, ArrayType
from numeta.array_shape import SCALAR


class ExternalLibraryWrapper(ExternalLibrary):
    """
    A wrapper class for external library.
    Used to convert types hint to fortran symbolic variables
    """

    def __init__(self, name, directory=None, include=None, additional_flags=None, to_link=True):
        super().__init__(name, directory, include, additional_flags, to_link=to_link)
        self.methods = ExternalModule(name, self, hidden=True)
        self.modules = {}

    def add_module(self, name, hidden=False):
        self.modules[name] = ExternalModule(name, self, hidden=False)

    def add_method(self, name, argtypes, restype, bind_c=True):
        symbolic_arguments = [
            convert_argument(f"a{i}", arg, bind_c=bind_c) for i, arg in enumerate(argtypes)
        ]
        return_type = None
        if restype is not None:
            return_type = convert_argument("res0", restype, bind_c=bind_c)

        self.methods.add_method(name, symbolic_arguments, return_type, bind_c=bind_c)

    def __getattr__(self, name):
        try:
            if name in self.methods.subroutines:
                return self.methods.subroutines[name]
            elif name in self.modules:
                return self.modules[name]
            else:
                raise AttributeError(f"Module {self.name} has no attribute {name}")
        except KeyError:
            raise AttributeError(f"ExternalLibrary object has no module {name}")


def convert_argument(name, hint, bind_c=True):
    if isinstance(hint, ArrayType):
        dtype = hint.dtype
        ftype = dtype.get_fortran(bind_c=bind_c)
        shape = hint.shape
    elif isinstance(hint, FortranType):
        ftype = hint
        shape = SCALAR
    elif isinstance(hint, type) and issubclass(hint, DataType):
        ftype = hint.get_fortran(bind_c=bind_c)
        shape = SCALAR
    elif isinstance(hint, type) and DataType.is_np_dtype(hint):
        dtype = DataType.from_np_dtype(hint)
        ftype = dtype.get_fortran(bind_c=bind_c)
        shape = SCALAR
    else:
        raise TypeError(f"Expected a numpy or numeta dtype got {type(hint).__name__}")

    return Variable(name, ftype=ftype, shape=shape)



import numeta as nm
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_scalar(dtype):
    @nm.jit
    def fill(a):
        tmp = nm.scalar(dtype, 50)
        a[0] = tmp

        tmp2 = nm.scalar(dtype, 100)
        a[1] = tmp2

    a = np.empty(2).astype(dtype)
    fill(a)

    np.testing.assert_allclose(a, [50, 100])



from numeta.syntax import PointerAssignment
from numeta.builder_helper import BuilderHelper
from numeta.array_shape import ArrayShape


def reshape(variable, shape, order="C"):
    if order not in ["C", "F"]:
        raise ValueError(f"Invalid order: {order}, must be 'C' or 'F'")
    fortran_order = order == "F"

    if not isinstance(shape, (tuple, list)):
        shape = (shape,)
    shape = ArrayShape(shape, fortran_order=fortran_order)

    builder = BuilderHelper.get_current_builder()
    pointer = builder.generate_local_variables(
        "fc_v",
        ftype=variable._ftype,
        shape=ArrayShape(tuple([None] * shape.rank), fortran_order=fortran_order),
        pointer=True,
    )

    PointerAssignment(pointer, shape, variable)

    return pointer



import numeta as nm
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_matmul_return_1_ndarray(dtype):
    n = 100

    @nm.jit
    def mul(a, b):
        c = nm.zeros((a.shape[0], b.shape[1]), dtype)
        for i in nm.range(a.shape[0]):
            for k in nm.range(b.shape[0]):
                c[i, :] += a[i, k] * b[k, :]
        return c

    a = np.random.rand(n, n).astype(dtype)
    b = np.random.rand(n, n).astype(dtype)

    c = mul(a, b)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(c, np.dot(a, b), atol=0)
    else:
        np.testing.assert_allclose(c, np.dot(a, b), rtol=10e2 * np.finfo(dtype).eps)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_matmul_return_2_ndarray(dtype):
    n = 100

    @nm.jit
    def mul(a, b):
        c = nm.zeros((a.shape[0], b.shape[1]), dtype)
        d = nm.zeros((a.shape[0], b.shape[1]), dtype)
        for i in nm.range(a.shape[0]):
            for k in nm.range(b.shape[0]):
                c[i, :] += a[i, k] * b[k, :]
        return c, d

    a = np.random.rand(n, n).astype(dtype)
    b = np.random.rand(n, n).astype(dtype)

    c, d = mul(a, b)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(c, np.dot(a, b), atol=0)
        np.testing.assert_allclose(d, np.zeros((n, n), dtype=dtype), atol=0)
    else:
        np.testing.assert_allclose(c, np.dot(a, b), rtol=10e2 * np.finfo(dtype).eps)
        np.testing.assert_allclose(
            d, np.zeros((n, n), dtype=dtype), rtol=10e2 * np.finfo(dtype).eps
        )


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_return_scalar(dtype):

    @nm.jit
    def return_scalar():
        return nm.scalar(dtype, 42)

    scalar = return_scalar()

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(scalar, 42, atol=0)
    else:
        np.testing.assert_allclose(scalar, 42, rtol=10e2 * np.finfo(dtype).eps)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_return_1_ndarray_getitem(dtype):
    n = 100
    m = 50

    @nm.jit
    def mul(a, b):
        c = nm.zeros((a.shape[0], b.shape[1]), dtype)
        for i in nm.range(a.shape[0]):
            for k in nm.range(b.shape[0]):
                c[i, :] += a[i, k] * b[k, :]
        return c[:10, : n // 2]

    a = np.random.rand(n, m).astype(dtype)
    b = np.random.rand(m, n).astype(dtype)

    c = mul(a, b)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(c, np.dot(a, b)[:10, : n // 2], atol=0)
    else:
        np.testing.assert_allclose(c, np.dot(a, b)[:10, : n // 2], rtol=10e2 * np.finfo(dtype).eps)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_return_1_ndarray_sum(dtype):
    n = 100
    m = 50

    @nm.jit
    def return_1_ndarray_sum(a, b):
        return a + b

    a = np.random.rand(n, m).astype(dtype)
    b = np.random.rand(n, m).astype(dtype)

    c = return_1_ndarray_sum(a, b)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(c, a + b, atol=0)
    else:
        np.testing.assert_allclose(c, a + b, rtol=10e2 * np.finfo(dtype).eps)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_return_1_ndarray_transpose(dtype):
    shape = (18, 7)

    @nm.jit
    def transpose_expr(a):
        return nm.transpose(a)

    a = np.random.random(shape).astype(dtype)

    out = transpose_expr(a)
    expected = np.transpose(a)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_array_equal(out, expected)
    else:
        np.testing.assert_allclose(out, expected, rtol=10e2 * np.finfo(dtype).eps)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_return_1_ndarray_mixed_expression(dtype):
    shape = (24, 16)

    @nm.jit
    def combined_expression(a, b):
        rows = a.shape[0] // 2
        cols = a.shape[1] // 2
        left = a[:rows, :cols] + b[:rows, :cols]
        right = a[:rows, :cols] - b[:rows, :cols]
        return left * right

    a = np.random.random(shape).astype(dtype)
    b = np.random.random(shape).astype(dtype)

    out = combined_expression(a, b)
    rows = shape[0] // 2
    cols = shape[1] // 2
    expected = (a[:rows, :cols] + b[:rows, :cols]) * (a[:rows, :cols] - b[:rows, :cols])

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_array_equal(out, expected)
    else:
        np.testing.assert_allclose(out, expected, rtol=10e2 * np.finfo(dtype).eps)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_call_returning_function_scalar(dtype):

    @nm.jit
    def return_scalar():
        return nm.scalar(dtype, 42)

    @nm.jit
    def caller():
        return return_scalar()

    scalar = caller()

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(scalar, 42, atol=0)
    else:
        np.testing.assert_allclose(scalar, 42, rtol=10e2 * np.finfo(dtype).eps)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_call_returning_function_array_rank1(dtype):

    size = 40

    @nm.jit
    def return_array_rank1():
        a = nm.zeros(size, dtype)
        return a

    @nm.jit
    def caller():
        return return_array_rank1()

    array = caller()

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(array, np.zeros(size, dtype=dtype), atol=0)
    else:
        np.testing.assert_allclose(
            array, np.zeros(size, dtype=dtype), rtol=10e2 * np.finfo(dtype).eps
        )


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_call_returning_function_array_rank2(dtype):

    @nm.jit
    def return_array_rank2(n, m):
        a = nm.zeros((n, m), dtype)
        return a

    @nm.jit
    def caller_rank2(n, m):
        return return_array_rank2(n, m)

    size = (10, 20)
    array = caller_rank2(*size)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(array, np.zeros(size, dtype=dtype), atol=0)
    else:
        np.testing.assert_allclose(
            array, np.zeros(size, dtype=dtype), rtol=10e2 * np.finfo(dtype).eps
        )


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_nested_return_array_rank1(dtype):

    size = 40

    @nm.jit
    def return_array_rank1():
        a = nm.zeros(size, dtype)
        return a

    @nm.jit
    def caller(b):
        b[:] = return_array_rank1()

    array = np.ones(size, dtype=dtype)
    caller(array)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(array, np.zeros(size, dtype=dtype), atol=0)
    else:
        np.testing.assert_allclose(
            array, np.zeros(size, dtype=dtype), rtol=10e2 * np.finfo(dtype).eps
        )


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_nested_return_array_rank1_no_numpy_allocator(dtype):

    size = 40

    nm.settings.unset_numpy_allocator()

    @nm.jit
    def return_array_rank1():
        a = nm.zeros(size, dtype)
        return a

    @nm.jit
    def caller(b):
        b[:] = return_array_rank1()

    array = np.ones(size, dtype=dtype)
    caller(array)

    nm.settings.set_numpy_allocator()

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(array, np.zeros(size, dtype=dtype), atol=0)
    else:
        np.testing.assert_allclose(
            array, np.zeros(size, dtype=dtype), rtol=10e2 * np.finfo(dtype).eps
        )


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_inline_returning_function_scalar(dtype):
    @nm.jit(inline=True)
    def return_scalar_inline():
        return nm.scalar(dtype, 7)

    @nm.jit
    def caller_scalar():
        return return_scalar_inline()

    scalar = caller_scalar()

    expected_scalar = dtype(7)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(scalar, expected_scalar, atol=0)
    else:
        np.testing.assert_allclose(scalar, expected_scalar, rtol=10e2 * np.finfo(dtype).eps)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_inline_returning_function_array(dtype):
    size = 8

    @nm.jit(inline=True)
    def return_array_inline(n):
        out = nm.zeros(n, dtype)
        out[:] = nm.scalar(dtype, 3)
        return out

    @nm.jit
    def caller_array(n):
        return return_array_inline(n)

    array = caller_array(size)

    expected_array = np.full(size, dtype(3), dtype=dtype)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(array, expected_array, atol=0)
    else:
        np.testing.assert_allclose(array, expected_array, rtol=10e2 * np.finfo(dtype).eps)



from .nodes import NamedEntity
from .scope import Scope
from .statements import SubroutineDeclaration, InterfaceDeclaration, Call
from .settings import settings


class Subroutine(NamedEntity):
    @classmethod
    def translate(cls, original_method):
        """Permit to translate the subroutine currently being generated."""

        def wrapper(self, *args, **kwargs):
            # Save the old scope, needed to restore it after the translation when constructing more subroutines
            old_scope = Scope.current_scope
            self.scope.enter()
            result = original_method(self, *args, **kwargs)
            self.scope.exit()
            Scope.current_scope = old_scope
            return result

        return wrapper

    def __init__(
        self,
        name,
        description=None,
        pure=False,
        elemental=False,
        to_print=True,
        parent=None,
        bind_c=None,
    ):
        super().__init__(name, parent=parent)
        self.description = description
        self.pure = pure
        self.elemental = elemental
        self.to_print = to_print
        self.arguments = {}  # dictionary because we need an ordered set
        self.scope = Scope()
        self.bind_c = settings.subroutine_bind_c if bind_c is None else bind_c
        self.declaration = None

        from .module import Module

        if isinstance(self.parent, Module):
            self.parent.add_subroutine(self)

    def add_variable(self, *variables, with_intent=None):
        """
        Add a variable to the subroutine. If the variable is a list, it is added as a list of variables.
        Can modify the intent of the variable if with_intent is specified.
        """
        for variable in variables:
            if type(variable) is list:
                for v in variable:
                    if with_intent is not None:
                        v.intent = with_intent
                    self.arguments[v.name] = v
            else:
                if with_intent is not None:
                    variable.intent = with_intent
                self.arguments[variable.name] = variable

    def add_to_description(self, value):
        self.description += value

    def get_dependencies(self):
        return self.get_declaration().dependencies

    def get_local_variables(self):
        return self.get_declaration().local_variables

    def get_external_interfaces(self):
        return self.get_declaration().interfaces

    def get_declaration(self):
        if self.declaration is None:
            self.declaration = SubroutineDeclaration(self)
        return self.declaration

    def count_statements(self):
        """Return the number of statements in this subroutine."""
        return sum(stmt.count_statements() for stmt in self.scope.get_statements())

    def print_lines(self, indent=0):
        return self.get_declaration().print_lines(indent=indent)

    def get_code(self):
        return "".join(self.print_lines())

    def get_interface_declaration(self):
        return InterfaceDeclaration(self)

    def get_interface_code(self):
        return "".join(self.get_interface_declaration().print_lines())

    def __call__(self, *arguments):
        Call(self, *arguments)



import numeta as nm
import numpy as np


def test_comptime_type_hint():

    n = 100
    m = 3

    @nm.jit
    def sum_first_n(length: nm.comptime, a, result):
        result[:] = 0.0
        for i in range(length):
            result[:] += a[i]

    array = np.random.random((n,))
    result = np.zeros((1,), dtype=array.dtype)

    for i in range(m):
        sum_first_n(i, array, result)
        np.testing.assert_allclose(result, array[:i].sum(), rtol=10e2 * np.finfo(array.dtype).eps)



from .statement import StatementWithScope
from .various import Comment, Use, Implicit, Contains
from .tools import (
    get_nested_dependencies_or_declarations,
    divide_variables_and_derived_types,
)


class ModuleDeclaration(StatementWithScope):
    def __init__(self, module):
        self.module = module

        entities = list(self.module.variables.values())

        dependencies, declarations = get_nested_dependencies_or_declarations(
            entities, self.module, for_module=True
        )
        self.variables_dec, self.derived_types_dec, functions_dec = (
            divide_variables_and_derived_types(declarations)
        )

        self.interfaces = [dec.subroutine for dec in functions_dec.values()]

        self.dependencies = {}
        self.modules_to_import = []

        from numeta.syntax.module import Module

        for dependency, var in dependencies:
            if hasattr(var, "get_interface_declaration"):
                # Should we add the interface?
                # Only if it is not contained in a module, if not the module will take care
                if not isinstance(dependency, Module) or dependency.hidden:
                    self.interfaces.append(var)
            if isinstance(dependency, Module):
                if not dependency.hidden:
                    self.modules_to_import.append((dependency, var))
                if dependency.parent is not None:
                    self.dependencies[dependency.parent.name] = dependency.parent
            else:
                self.dependencies[dependency.name] = dependency

    @property
    def children(self):
        return []

    def extract_entities(self):
        # Assume nothing is visible outside the Subroutine (maybe not okay?)
        yield self.module

    def get_statements(self):
        if self.module.description is not None:
            for line in self.module.description.split("\n"):
                yield Comment(line, add_to_scope=False)

        for dependency, variable in self.modules_to_import:
            yield Use(dependency, only=variable, add_to_scope=False)

        yield from self.derived_types_dec.values()

        if self.interfaces:
            raise NotImplementedError("Interfaces are not supported yet")

        yield Implicit(implicit_type="none", add_to_scope=False)

        yield from self.variables_dec.values()

        yield Contains(add_to_scope=False)

        yield from self.module.subroutines.values()

    def get_start_code_blocks(self):
        return ["module", " ", self.module.name]

    def get_end_code_blocks(self):
        return ["end", " ", "module", " ", self.module.name]



from .expression_node import ExpressionNode
from numeta.syntax.nodes import NamedEntity
from numeta.syntax.tools import check_node


class Function(NamedEntity, ExpressionNode):
    __slots__ = ["name", "arguments"]

    def __init__(self, name, arguments, parent=None):
        super().__init__(name, parent=parent)
        self.arguments = [check_node(arg) for arg in arguments]

    def get_code_blocks(self):
        result = [self.name, "("]
        for argument in self.arguments:
            result.extend(argument.get_code_blocks())
            result.append(", ")
        if result[-1] == ", ":
            result.pop()
        result.append(")")
        return result

    def extract_entities(self):
        yield self
        for arg in self.arguments:
            yield from arg.extract_entities()

    def get_declaration(self):
        raise NotImplementedError("Function declaration is not supported")

    def get_with_updated_variables(self, variables_couples):
        new_args = [arg.get_with_updated_variables(variables_couples) for arg in self.arguments]
        return type(self)(self.name, new_args, parent=self.parent)



from .function import Function
from numeta.syntax.tools import check_node
from numeta.syntax.settings import settings
from numeta.array_shape import ArrayShape, SCALAR, UNKNOWN


class IntrinsicFunction(Function):
    token = ""

    def __init__(self, *arguments):
        from numeta.syntax.module import builtins_module

        super().__init__(self.token, [check_node(arg) for arg in arguments], parent=builtins_module)

    def get_with_updated_variables(self, variables_couples):
        new_args = [arg.get_with_updated_variables(variables_couples) for arg in self.arguments]
        return type(self)(*new_args)

    @property
    def _ftype(self):
        # default behavior for a lot of intrinsic functions
        return self.arguments[0]._ftype

    @property
    def _shape(self):
        # default behavior for a lot of intrinsic functions
        return self.arguments[0]._shape


class UnaryIntrinsicFunction(IntrinsicFunction):

    def __init__(self, argument):
        super().__init__(check_node(argument))


class BinaryIntrinsicFunction(IntrinsicFunction):

    def __init__(self, argument1, argument2):
        super().__init__(
            check_node(argument1),
            check_node(argument2),
        )


class Abs(UnaryIntrinsicFunction):
    token = "abs"


class Neg(UnaryIntrinsicFunction):
    token = "-"


class Not(UnaryIntrinsicFunction):
    token = ".not."


class Allocated(UnaryIntrinsicFunction):
    token = "allocated"

    @property
    def _ftype(self):
        return settings.DEFAULT_LOGICAL

    @property
    def _shape(self):
        return SCALAR


class All(UnaryIntrinsicFunction):
    token = "all"

    @property
    def _ftype(self):
        return settings.DEFAULT_LOGICAL

    @property
    def _shape(self):
        return SCALAR


class Shape(UnaryIntrinsicFunction):
    token = "shape"

    def __init__(self, argument):
        if argument._shape is SCALAR:
            raise ValueError("The shape intrinsic function cannot be applied to a scalar.")
        super().__init__(argument)

    @property
    def _ftype(self):
        return self.arguments[0]._ftype

    @property
    def _shape(self):
        var_shape = self.arguments[0]._shape
        if var_shape is SCALAR or var_shape is UNKNOWN:
            raise ValueError(
                "The shape intrinsic function can only be applied to variables with a defined shape."
            )
        return ArrayShape((len(var_shape.dims),))


class Real(UnaryIntrinsicFunction):
    token = "real"

    @property
    def _ftype(self):
        return settings.DEFAULT_REAL


class Imag(UnaryIntrinsicFunction):
    token = "aimag"

    @property
    def _ftype(self):
        return settings.DEFAULT_REAL


class Conjugate(UnaryIntrinsicFunction):
    token = "conjg"

    @property
    def _ftype(self):
        return settings.DEFAULT_COMPLEX

    @property
    def _shape(self):
        return self.arguments[0]._shape


class Complex(Function):
    def __init__(self, real, imaginary, kind=None):
        self.name = "cmplx"
        if kind is None:
            kind = settings.DEFAULT_COMPLEX.kind
        self.arguments = [check_node(real), check_node(imaginary), check_node(kind)]

    @property
    def _ftype(self):
        # TODO to fix, not consistent with Optional kind
        return settings.DEFAULT_COMPLEX

    @property
    def _shape(self):
        return self.arguments[0]._shape


class Transpose(UnaryIntrinsicFunction):
    token = "transpose"

    @property
    def shape(self):
        if self.arguments[0]._shape is SCALAR:
            raise ValueError("Cannot transpose a scalar.")
        elif self.arguments[0]._shape is UNKNOWN:
            raise ValueError("Cannot transpose a variable with unknown shape.")
        elif len(self.arguments[0]._shape.dims) != 2:
            raise ValueError("Transpose can only be applied to 2-D arrays.")
        return ArrayShape(self.arguments[0]._shape.dims[::-1])


class Exp(UnaryIntrinsicFunction):
    token = "exp"


class Sqrt(UnaryIntrinsicFunction):
    token = "sqrt"


class Floor(UnaryIntrinsicFunction):
    token = "floor"


class Sin(UnaryIntrinsicFunction):
    token = "sin"


class Cos(UnaryIntrinsicFunction):
    token = "cos"


class Tan(UnaryIntrinsicFunction):
    token = "tan"


class Sinh(UnaryIntrinsicFunction):
    token = "sinh"


class Cosh(UnaryIntrinsicFunction):
    token = "cosh"


class Tanh(UnaryIntrinsicFunction):
    token = "tanh"


class ASin(UnaryIntrinsicFunction):
    token = "asin"


class ACos(UnaryIntrinsicFunction):
    token = "acos"


class ATan(UnaryIntrinsicFunction):
    token = "atan"


class ATan2(BinaryIntrinsicFunction):
    token = "atan2"


class Dotproduct(BinaryIntrinsicFunction):
    token = "dot_product"

    @property
    def _ftype(self):
        return self.arguments[0]._ftype

    @property
    def _shape(self):
        return SCALAR


class Rank(UnaryIntrinsicFunction):
    token = "rank"

    @property
    def _ftype(self):
        return settings.DEFAULT_INTEGER

    @property
    def _shape(self):
        return SCALAR


class Size(BinaryIntrinsicFunction):
    token = "size"

    @property
    def _ftype(self):
        return settings.DEFAULT_INTEGER

    @property
    def _shape(self):
        return SCALAR


class Max(IntrinsicFunction):
    token = "max"

    @property
    def _ftype(self):
        return self.arguments[0]._ftype

    @property
    def _shape(self):
        return SCALAR


class Maxval(UnaryIntrinsicFunction):
    token = "maxval"

    @property
    def _ftype(self):
        return self.arguments[0]._ftype

    @property
    def _shape(self):
        return SCALAR


class Min(IntrinsicFunction):
    token = "min"

    @property
    def _ftype(self):
        return self.arguments[0]._ftype

    @property
    def _shape(self):
        return SCALAR


class Minval(UnaryIntrinsicFunction):
    token = "minval"

    @property
    def _ftype(self):
        return self.arguments[0]._ftype

    @property
    def _shape(self):
        return SCALAR


class Iand(BinaryIntrinsicFunction):
    token = "iand"


class Ior(BinaryIntrinsicFunction):
    token = "ior"


class Xor(BinaryIntrinsicFunction):
    token = "xor"


class Ishft(BinaryIntrinsicFunction):
    token = "ishft"


class Ibset(BinaryIntrinsicFunction):
    token = "ibset"


class Ibclr(BinaryIntrinsicFunction):
    token = "ibclr"


class Popcnt(UnaryIntrinsicFunction):
    token = "popcnt"

    @property
    def _ftype(self):
        return settings.DEFAULT_INTEGER

    @property
    def _shape(self):
        return SCALAR


class Trailz(UnaryIntrinsicFunction):
    token = "trailz"

    @property
    def _ftype(self):
        return settings.DEFAULT_INTEGER

    @property
    def _shape(self):
        return SCALAR


class Sum(UnaryIntrinsicFunction):
    token = "sum"

    @property
    def _ftype(self):
        return self.arguments[0]._ftype

    @property
    def _shape(self):
        return SCALAR


class Matmul(BinaryIntrinsicFunction):
    token = "matmul"

    @property
    def _ftype(self):
        return self.arguments[0]._ftype

    @property
    def _shape(self):
        a_shape = self.arguments[0]._shape
        b_shape = self.arguments[1]._shape
        if len(a_shape.dims) == 1:
            return ArrayShape((b_shape.dims[1],))
        if len(b_shape.dims) == 1:
            return ArrayShape((a_shape.dims[0],))
        return ArrayShape((a_shape.dims[0], b_shape.dims[1]))



import numpy as np
import numeta as nm


def test_struct_array():
    n = 2
    m = 3

    np_nested1 = np.dtype([("a", np.int64, (n, n)), ("b", np.float64, (m,))], align=True)
    np_nested2 = np.dtype([("c", np_nested1, (n,)), ("d", np_nested1, (3,))], align=True)
    np_nested3 = np.dtype([("c", np_nested2, (2,)), ("d", np_nested1, (3,))], align=True)

    @nm.jit
    def mod_struct(a) -> None:
        a[1]["c"][1]["d"][2]["b"][1] = -4.0

    a = np.zeros(2, dtype=np_nested3)

    mod_struct(a)

    b = np.zeros(2, dtype=np_nested3)
    b[1]["c"][1]["d"][2]["b"][1] = -4.0

    np.testing.assert_equal(a, b)


def test_struct():
    n = 2
    m = 3

    np_nested1 = np.dtype([("a", np.int64, (n, n)), ("b", np.float64, (m,))], align=True)
    np_nested2 = np.dtype([("c", np_nested1, (n,)), ("d", np_nested1, (3,))], align=True)
    np_nested3 = np.dtype([("c", np_nested2, (2,)), ("d", np_nested1, (3,))], align=True)

    @nm.jit
    def mod_struct(a) -> None:
        a["c"][1]["d"][2]["b"][1] = -4.0

    a = np.zeros(2, dtype=np_nested3)

    mod_struct(a[1])

    b = np.zeros(2, dtype=np_nested3)
    b[1]["c"][1]["d"][2]["b"][1] = -4.0

    np.testing.assert_equal(a, b)


def test_struct_type_helpers():
    dtype = np.dtype([("x", np.int32), ("y", np.float64)], align=True)
    st = nm.get_datatype(dtype)

    arr_t = st[2]
    assert isinstance(arr_t, nm.datatype.ArrayType)
    assert arr_t.dtype is st
    assert arr_t.shape.dims == (2,)

    @nm.jit
    def fill(out):
        tmp = st()
        tmp["x"] = 5
        tmp["y"] = 1.5
        out[0] = tmp

    out = np.zeros(1, dtype=dtype)
    fill(out)

    expected = np.zeros(1, dtype=dtype)
    expected[0]["x"] = 5
    expected[0]["y"] = 1.5
    np.testing.assert_equal(out, expected)


def test_struct_array_call():
    dtype = np.dtype([("x", np.int32), ("y", np.float64)], align=True)
    nm_dtype = nm.get_datatype(dtype)

    @nm.jit
    def fill(out):
        val = nm_dtype()
        val["x"] = 7
        val["y"] = -2.0
        out[:] = nm_dtype[2](val)

    out = np.empty(2, dtype=dtype)
    fill(out)

    expected = np.empty(2, dtype=dtype)
    expected["x"] = 7
    expected["y"] = -2.0
    np.testing.assert_equal(out, expected)


def test_struct_class_reuse():

    dtype = np.dtype([("x", np.int32), ("y", np.float64)], align=True)
    st1 = nm.get_datatype(dtype)
    st2 = nm.get_datatype(dtype)
    assert st1 is st2



from numeta.syntax.nodes import Node
from numeta.syntax.scope import Scope
from .tools import print_block


class Statement(Node):
    """
    A simple statement that can be printed and executed within a code block.

    Methods
    -------
    extract_entities():
        Extract entities referenced within the statement that have to be defined outside.

    get_code_blocks():
        Return the code blocks representing the statement.

    print_lines(indent=0):
        Print the statement as a list of formatted strings.
    """

    def __init__(self, add_to_scope=True):
        if add_to_scope:
            Scope.add_to_current_scope(self)

    @property
    def children(self):
        """Return the child nodes of the statement."""
        raise NotImplementedError(
            f"Subclass '{self.__class__.__name__}' must implement the 'children' property."
        )

    def get_code_blocks(self):
        """Return the code blocks for this statement."""
        raise NotImplementedError(
            f"Subclass '{self.__class__.__name__}' must implement 'get_code_blocks'."
        )

    def extract_entities(self):
        """Recursively extract all entities within the statement's child nodes."""
        for child in self.children:
            yield from child.extract_entities()

    def print_lines(self, indent=0):
        """Print the statement, formatted with the given indent level."""
        return [print_block(self.get_code_blocks(), indent=indent)]

    def count_statements(self):
        """Return 1 for simple statements."""
        return 1

    def get_with_updated_variables(self, variables_couples):
        new_children = [
            child.get_with_updated_variables(variables_couples) for child in self.children
        ]
        return type(self)(*new_children, add_to_scope=False)


class StatementWithScope(Statement):
    """
    A statement that introduces a new scope (e.g., if, do, subroutine).

    Methods
    -------
    extract_entities():
        Extract entities referenced in the scope.

    get_start_code_blocks():
        Get the starting code blocks of the scope.

    get_statements():
        Get the list of statements within the scope.

    get_end_code_blocks():
        Get the ending code blocks of the scope.

    print_lines(indent=0):
        Print the statement, including its body, as a list of formatted strings.
    """

    def __init__(self, add_to_scope=True, enter_scope=True):
        super().__init__(add_to_scope=add_to_scope)
        self.scope = Scope()
        if enter_scope:
            self.scope.enter()

    @property
    def children(self):
        """Return the child nodes in the start code and end code block."""
        raise NotImplementedError(
            f"Subclass '{self.__class__.__name__}' must implement the 'children' property."
        )

    def get_start_code_blocks(self):
        """Return the code blocks that start this scoped statement."""
        raise NotImplementedError(
            f"Subclass '{self.__class__.__name__}' must implement 'get_start_code_blocks'."
        )

    def get_end_code_blocks(self):
        """Return the code blocks that end this scoped statement."""
        raise NotImplementedError(
            f"Subclass '{self.__class__.__name__}' must implement 'get_end_code_blocks'."
        )

    def __str__(self):
        """Return the string representation of the scoped statement."""
        return "".join(self.print_lines(indent=0))

    def get_statements(self):
        """Return the list of statements within the scope."""
        return self.scope.get_statements()

    def print_lines(self, indent=0):
        """Print the entire scoped statement, including all nested statements."""
        result = [print_block(self.get_start_code_blocks(), indent=indent)]
        for statement in self.get_statements():
            result.extend(statement.print_lines(indent=indent + 1))
        if len(last_statement := self.get_end_code_blocks()) > 0:
            result.append(print_block(last_statement, indent=indent))
        return result

    def extract_entities(self):
        """Extract all the visible from outside entities within this statement and its scoped statements."""
        for child in self.children:
            yield from child.extract_entities()
        for statement in self.get_statements():
            yield from statement.extract_entities()

    def get_with_updated_variables(self, variables_couples):
        new_children = [
            child.get_with_updated_variables(variables_couples) for child in self.children
        ]
        result = type(self)(*new_children, add_to_scope=False, enter_scope=False)
        result.scope = self.scope.get_with_updated_variables(variables_couples)

        return result

    def count_statements(self):
        """Count this statement and all nested statements."""
        count = 1
        for statement in self.get_statements():
            count += statement.count_statements()
        return count

    def __enter__(self):
        """Enter the scope of the statement."""
        # Add specific enter logic if needed
        pass

    def __exit__(self, *args):
        """Exit the scope of the statement."""
        self.scope.exit()



from .expression_node import ExpressionNode
from .binary_operation_node import BinaryOperationNode, EqBinaryNode, NeBinaryNode
from .getattr import GetAttr
from .getitem import GetItem
from .literal_node import LiteralNode
from .function import Function
from .intrinsic_functions import *
from .various import *



from .various import *
from .call import Call
from .variable_declaration import VariableDeclaration
from .derived_type_declaration import DerivedTypeDeclaration
from .subroutine_declaration import SubroutineDeclaration, InterfaceDeclaration
from .module_declaration import ModuleDeclaration



import numeta as nm
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_cast(dtype):

    @nm.jit
    def set_nine(a):
        a_int = nm.cast(a, dtype)
        a_int[:] = 9.0

    # should contain everything
    a = np.ones(16, dtype=np.bool_)
    set_nine(a)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(a.view(dtype)[0], np.array(9, dtype=dtype), atol=0)
    else:
        np.testing.assert_allclose(
            a.view(dtype)[0], np.array(9, dtype=dtype), rtol=10e2 * np.finfo(dtype).eps
        )


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_cast_getitem(dtype):

    @nm.jit
    def set_nine(a):
        a_int = nm.cast(a[0], dtype)
        a_int[:] = 9.0

    # should contain everything
    a = np.ones(16, dtype=np.bool_)
    set_nine(a)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(a.view(dtype)[0], np.array(9, dtype=dtype), atol=0)
    else:
        np.testing.assert_allclose(
            a.view(dtype)[0], np.array(9, dtype=dtype), rtol=10e2 * np.finfo(dtype).eps
        )


def test_cast_struct():

    dtype = np.dtype([("a", np.float64), ("b", np.int64)])

    @nm.jit
    def set_nine(a):
        a_int = nm.cast(a, dtype)
        a_int["a"][:] = 9.0
        a_int["b"][:] = 9

    # should contain everything
    a = np.ones(16, dtype=np.bool_)
    set_nine(a)

    check = np.empty((), dtype)
    check["a"] = 9.0
    check["b"] = 9

    print(a.view(dtype)["a"][0], check["a"])
    print(a.view(dtype)["b"][0], check["b"])

    # cannot use rtol
    np.testing.assert_allclose(a.view(dtype)["a"][0], check["a"], atol=0)
    np.testing.assert_allclose(a.view(dtype)["b"][0], check["b"], atol=0)



from .expression_node import ExpressionNode
from numeta.syntax.settings import settings
from numeta.array_shape import ArrayShape, UNKNOWN, SCALAR


class GetItem(ExpressionNode):
    def __init__(self, variable, slice_):
        self.variable = variable
        # define if only a slice [begin : end : step] of the Variable is asked
        self.sliced = slice_

    @property
    def target(self):
        return self.variable.target

    @target.setter
    def target(self, value):
        self.variable.target = value

    @property
    def _ftype(self):
        return self.variable._ftype

    @property
    def _shape(self):
        def get_dim_slice(slice_, max_dim):
            start = slice_.start
            if start is None:
                start = settings.array_lower_bound
            stop = slice_.stop
            if stop is None and max_dim is not None:
                stop = max_dim
            if slice_.step is not None:
                raise NotImplementedError("Step slicing not implemented for shape extraction")
            if stop is None:
                return None
            return stop - start + settings.array_lower_bound

        dims = []
        if self.variable._shape is UNKNOWN:
            if isinstance(self.sliced, slice):
                dims.append(get_dim_slice(self.sliced, None))
            else:
                dims.append(1)
        elif self.variable._shape is SCALAR:
            dims.append(1)
        elif isinstance(self.sliced, tuple):
            for i, element in enumerate(self.sliced):
                if isinstance(element, slice):
                    dims.append(get_dim_slice(element, self.variable._shape.dims[i]))
                else:
                    dims.append(1)
        else:
            if isinstance(self.sliced, slice):
                dims.append(get_dim_slice(self.sliced, self.variable._shape.dims[0]))
            else:
                dims.append(1)

        return ArrayShape(tuple(dims))

    def extract_entities(self):
        yield from self.variable.extract_entities()
        from numeta.syntax.tools import extract_entities

        yield from extract_entities(self.sliced)

    def get_code_blocks(self):
        result = self.variable.get_code_blocks()

        def get_block(block):
            if hasattr(block, "get_code_blocks"):
                return block.get_code_blocks()
            else:
                return [str(block)]

        def convert_slice(slice_):
            result = []

            if slice_.start is not None:
                result += get_block(slice_.start)

            result.append(":")

            if slice_.stop is not None:
                stop = slice_.stop - 1 if settings.c_like_bounds else slice_.stop
                result += get_block(stop)

            if slice_.step is not None:
                result.append(":")
                result += get_block(slice_.step)

            return result

        result.append("(")
        if isinstance(self.sliced, tuple):
            dims = []
            if hasattr(self.sliced[0], "get_code_blocks"):
                dims.append(self.sliced[0].get_code_blocks())
            elif isinstance(self.sliced[0], slice):
                dims.append(convert_slice(self.sliced[0]))
            else:
                dims.append([str(self.sliced[0])])
            for element in self.sliced[1:]:
                if hasattr(element, "get_code_blocks"):
                    dims.append(element.get_code_blocks())
                elif isinstance(element, slice):
                    dims.append(convert_slice(element))
                else:
                    dims.append([str(element)])
            if not self.variable._shape.fortran_order:
                dims = dims[::-1]

            result += dims[0]
            for dim in dims[1:]:
                result += [",", " "]
                result += dim

        else:
            if hasattr(self.sliced, "get_code_blocks"):
                result += self.sliced.get_code_blocks()
            elif isinstance(self.sliced, slice):
                result += convert_slice(self.sliced)
            else:
                result += [str(self.sliced)]
        result.append(")")
        return result

    def __setitem__(self, key, value):
        from numeta.syntax.statements import Assignment

        Assignment(self[key], value)

    def get_with_updated_variables(self, variables_couples):

        from numeta.syntax.tools import update_variables

        new_var = self.variable.get_with_updated_variables(variables_couples)
        new_slice = update_variables(self.sliced, variables_couples)

        # If the variable was replaced by another GetItem (e.g. during inlining),
        # compose the indexing using the standard __getitem__ logic so that
        # slices are merged correctly.
        if isinstance(new_var, GetItem):
            return new_var[new_slice]

        # If the variable is an ArrayConstructor and the slice is an int,
        # HACK because fortran does not treat temporary variables as first class citizens
        # only for ArrayConstructor
        from .various import ArrayConstructor

        if isinstance(new_var, ArrayConstructor) and isinstance(self.sliced, int):
            return new_var.elements[self.sliced]

        return GetItem(new_var, new_slice)

    def __getitem__(self, key):
        if isinstance(key, str):
            from .getattr import GetAttr

            return GetAttr(self, key)
        else:
            new_key = self.merge_slice(key)
            return GetItem(self.variable, new_key)

    def merge_slice(self, key):
        """
        Merge the slice key with the current slice
        So for example:

            a[5:10][2:4] -> a[6:8]
        """
        if isinstance(self.sliced, slice):
            if key is None:
                new_key = self.sliced

            elif isinstance(key, slice):
                if self.sliced.step is not None or key.step is not None:
                    raise Warning("Step slicing not implemented for slice merging")

                lb = settings.array_lower_bound

                if self.sliced.start is None and self.sliced.stop is None:
                    new_start = key.start
                    new_stop = key.stop
                else:
                    base_start = self.sliced.start if self.sliced.start is not None else lb

                    if key.start is None:
                        new_start = self.sliced.start
                    elif self.sliced.start is None:
                        new_start = key.start
                    else:
                        new_start = base_start + key.start - lb

                    if key.stop is None:
                        new_stop = self.sliced.stop
                    elif self.sliced.start is None:
                        new_stop = key.stop
                    else:
                        new_stop = base_start + key.stop - lb

                new_key = slice(new_start, new_stop, None)

            else:
                lb = settings.array_lower_bound
                base_start = self.sliced.start if self.sliced.start is not None else lb
                new_key = base_start + key - lb

        elif key is None:
            new_key = self.sliced

        else:
            error_str = "Error in array slicing. Cannot merge old slice with new one.\n"
            error_str += f"\nName of the variable: {self.variable.name}"
            error_str += f"\nOld slice: {self.sliced}"
            error_str += f"\nNew slice: {key}"
            error_str += f"\nImpossible to merge {self.variable.name}[{self.sliced}][{key}]"
            raise Warning(error_str)

        return new_key



from __future__ import annotations

from pathlib import Path
import subprocess as sp
import textwrap
from typing import Iterable


class FortranCompiler:
    def __init__(self, compile_flags: str | Iterable[str]) -> None:
        self.compile_flags = self._normalize_flags(compile_flags)

    @staticmethod
    def _normalize_flags(compile_flags: str | Iterable[str]) -> list[str]:
        if isinstance(compile_flags, str):
            return compile_flags.split()
        return list(compile_flags)

    def build_fortran_command(
        self,
        *,
        obj_file: Path,
        source_file: Path,
        include_dirs: Iterable[str],
        additional_flags: Iterable[str],
    ) -> list[str]:
        command = ["gfortran"]
        command.extend(["-fopenmp"])
        command.extend(self.compile_flags)
        command.extend(["-fPIC", "-c", "-o", str(obj_file)])
        command.append(str(source_file))
        command.extend([f"-I{inc_dir}" for inc_dir in include_dirs])
        command.extend(additional_flags)
        return command

    def run_command(self, command: list[str], cwd: Path) -> sp.CompletedProcess[str]:
        sp_run = sp.run(
            command,
            cwd=cwd,
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            text=True,
        )
        if sp_run.returncode != 0:
            error_message = "Error while compiling, the command was:\n"
            error_message += " ".join(command) + "\n"
            error_message += "The output was:\n"
            error_message += textwrap.indent(sp_run.stdout, "    ")
            error_message += textwrap.indent(sp_run.stderr, "    ")
            raise Warning(error_message)
        return sp_run

    def compile_fortran(
        self,
        *,
        name: str,
        directory: Path,
        source: str,
        dependencies: Iterable[object],
    ) -> tuple[Path, str]:
        """
        Compile Fortran source files using gfortran and return the resulting object file.
        """
        fortran_src = directory / f"{name}_src.f90"
        fortran_src.write_text(source)

        obj_file = directory / f"{name}_fortran.o"

        include_dirs: list[str] = []
        additional_flags: list[str] = []

        for lib in dependencies:
            if lib.include is not None:
                if isinstance(lib.include, (list, tuple, set)):
                    include_dirs.extend(lib.include)
                else:
                    include_dirs.append(lib.include)
            if lib.additional_flags is not None:
                if isinstance(lib.additional_flags, str):
                    additional_flags.extend(lib.additional_flags.split())
                else:
                    additional_flags.append(lib.additional_flags)

        command = self.build_fortran_command(
            obj_file=obj_file,
            source_file=fortran_src,
            include_dirs=include_dirs,
            additional_flags=additional_flags,
        )

        self.run_command(command, cwd=directory)

        return obj_file, str(directory)



from .syntax import Variable, Scope, Subroutine
from .settings import settings


class BuilderHelper:
    current_builder = None

    @classmethod
    def set_current_builder(cls, builder):
        cls.current_builder = builder

    @classmethod
    def get_current_builder(cls):
        if cls.current_builder is None:
            raise Warning("The current builder is not initialized")
        return cls.current_builder

    def __init__(self, numeta_function, symbolic_function, signature):
        self.numeta_function = numeta_function
        self.symbolic_function = symbolic_function
        self.signature = signature

        self.prefix_counter = {}
        self.allocated_arrays = {}

        if settings.use_numpy_allocator:
            self.allocate_array = self._allocate_array_numpy
            self.deallocate_array = self._deallocate_array_numpy
        else:
            self.allocate_array = self._allocate_array
            self.deallocate_array = self._deallocate_array

    @classmethod
    def generate_local_variables(cls, prefix, allocate=False, name=None, **kwargs):
        """
        TODO:
        TO DEPRECATE in some way, use:
        builder = BuilderHelper.get_current_builder()
        bilder.generate_local_variable(...)
        problem: sometimes is not required to have a builder (i.e. fixed name) so maybe set a default one (?).
        """
        if name is None:
            builder = cls.get_current_builder()
            if prefix not in builder.prefix_counter:
                builder.prefix_counter[prefix] = 0
            builder.prefix_counter[prefix] += 1
            name = f"{prefix}{builder.prefix_counter[prefix]}"
        if allocate:
            builder = cls.get_current_builder()
            return builder.allocate_array(name, **kwargs)
        return Variable(name, **kwargs)

    def generate_local_variable(self, prefix, allocate=False, name=None, **kwargs):
        return BuilderHelper.generate_local_variables(
            prefix, allocate=allocate, name=name, **kwargs
        )

    def _allocate_array(self, name, shape, **kwargs):
        from .syntax import Allocate, If, Allocated, Not
        from .array_shape import ArrayShape

        alloc_shape = ArrayShape(tuple([None] * shape.rank), fortran_order=shape.fortran_order)
        variable = Variable(name, shape=alloc_shape, allocatable=True, **kwargs)
        with If(Not(Allocated(variable))):
            Allocate(variable, *shape.dims)
        self.allocated_arrays[name] = variable
        return variable

    def _deallocate_array(self, array):
        from numeta.syntax import Deallocate, If, Allocated

        with If(Allocated(array)):
            Deallocate(array)

    def _allocate_array_numpy(self, name, shape, **kwargs):
        from .syntax import PointerAssignment
        from .syntax.expressions import ArrayConstructor
        from .wrappers import numpy_mem
        from .external_modules.iso_c_binding import FPointer_c, iso_c
        from .array_shape import ArrayShape
        from .datatype import DataType

        # create a c pointer variable that will be also deallocated
        variable_ptr = Variable(f"{name}_c_ptr", FPointer_c)
        self.allocated_arrays[name] = variable_ptr

        dtype = DataType.from_ftype(kwargs["ftype"])

        size = dtype.get_nbytes()
        for dim in shape.dims:
            size *= dim

        # allocate memory with the numpy allocator
        numpy_mem.numpy_allocate(variable_ptr, size)

        # Fortran is so versone
        # create fortran pointer (with lower bound 1)
        variable_lb1 = Variable(
            f"{name}_f_ptr_lb1", ftype=kwargs["ftype"], shape=ArrayShape((None,)), pointer=True
        )
        # point the fortran pointer to the allocated memory
        iso_c.c_f_pointer(variable_ptr, variable_lb1, ArrayConstructor(size))

        alloc_shape = ArrayShape(tuple([None] * shape.rank), fortran_order=shape.fortran_order)
        variable = Variable(name, shape=alloc_shape, pointer=True, **kwargs)

        # assign the fortran pointer with the proper lower bound
        PointerAssignment(variable, shape, variable_lb1)

        return variable

    def _deallocate_array_numpy(self, array):
        from .wrappers import numpy_mem

        numpy_mem.numpy_deallocate(array)

    def build(self, *args, **kwargs):
        old_builder = self.current_builder
        self.set_current_builder(self)

        old_scope = Scope.current_scope
        self.symbolic_function.scope.enter()

        return_variables = self.numeta_function.run_symbolic(*args, **kwargs)

        if return_variables is None:
            return_variables = []
        elif not isinstance(return_variables, (list, tuple)):
            return_variables = [return_variables]

        from .array_shape import ArrayShape, SCALAR, UNKNOWN
        from .syntax import Shape
        from .datatype import DataType, size_t

        ret = []
        for i, var in enumerate(return_variables):
            expr = None
            if isinstance(var, Variable):
                if var.name in self.allocated_arrays:

                    rank = var._shape.rank
                    shape = Variable(
                        f"fc_out_shape_{i}",
                        ftype=size_t.get_fortran(bind_c=True),
                        shape=ArrayShape((rank,)),
                        intent="out",
                    )
                    self.symbolic_function.add_variable(shape)
                    # add to the symbolic function
                    shape[:] = Shape(var)
                    shape[:] = shape[rank - 1 : 1 : -1]  # reverse the shape for Fortran order

                    ptr = self.allocated_arrays.pop(var.name)
                    ptr.intent = "out"
                    self.symbolic_function.add_variable(ptr)

                    ret.append((DataType.from_ftype(var._ftype), rank))
                elif var._shape is SCALAR and var.name not in self.symbolic_function.arguments:

                    var.intent = "out"
                    self.symbolic_function.add_variable(var)

                    ret.append((DataType.from_ftype(var._ftype), 0))
                else:
                    if var._shape is SCALAR:
                        tmp = BuilderHelper.generate_local_variables("fc_s", ftype=var._ftype)
                        tmp[:] = var
                        tmp.intent = "out"
                        self.symbolic_function.add_variable(tmp)
                        ret.append((DataType.from_ftype(var._ftype), 0))
                        continue
                    expr = var
            else:
                # it is an expression
                expr = var

            if expr is not None:
                # We have to copy the expression in a new array
                expr_shape = expr._shape
                if expr_shape is SCALAR:
                    tmp = BuilderHelper.generate_local_variables("fc_s", ftype=expr._ftype)
                    tmp[:] = expr
                    tmp.intent = "out"
                    self.symbolic_function.add_variable(tmp)
                    ret.append((DataType.from_ftype(expr._ftype), 0))
                    continue

                if expr_shape is UNKNOWN:
                    raise NotImplementedError(
                        "Returning arrays with unknown shape is not supported yet."
                    )

                rank = expr_shape.rank
                shape = Variable(
                    f"fc_out_shape_{i}",
                    ftype=size_t.get_fortran(bind_c=True),
                    shape=ArrayShape((rank,)),
                    intent="out",
                )
                self.symbolic_function.add_variable(shape)

                # Cache the expression shape in a local array so we can reuse it
                # both for allocating the temporary buffer and to report the
                # extents back through the C API.
                tmp_shape = Variable(
                    f"fc_out_shape_{i}_tmp",
                    ftype=size_t.get_fortran(bind_c=True),
                    shape=ArrayShape((rank,)),
                )
                tmp_shape[:] = Shape(expr)

                alloc_dims = [tmp_shape[i] for i in range(rank)]
                if not expr_shape.fortran_order:
                    alloc_dims = alloc_dims[::-1]

                from .wrappers import empty

                tmp = empty(
                    alloc_dims,
                    dtype=expr._ftype,
                    order="F" if expr_shape.fortran_order else "C",
                )

                shape[:] = tmp_shape
                # Reverse the shape for the C interface after materializing the
                # temporary buffer so the allocation happens with the original
                # extents.
                shape[:] = shape[rank - 1 : 1 : -1]
                tmp[:] = expr

                ptr = self.allocated_arrays.pop(tmp.name)
                ptr.intent = "out"
                self.symbolic_function.add_variable(ptr)

                ret.append((DataType.from_ftype(expr._ftype), rank))

        for array in self.allocated_arrays.values():
            self.deallocate_array(array)

        self.symbolic_function.scope.exit()
        Scope.current_scope = old_scope

        self.set_current_builder(old_builder)

        return ret

    def inline(self, function, *arguments):
        """Inline ``function`` with the given ``arguments`` into the current scope."""
        # Avoid heavy imports at module load time
        if not isinstance(function, Subroutine):
            raise TypeError("Unsupported function type for inline call")

        from .syntax.tools import check_node

        args = [check_node(arg) for arg in arguments]
        if len(args) != len(function.arguments):
            raise ValueError("Incorrect number of arguments for inlined subroutine")
        variables_couples = list(zip(function.arguments.values(), args))
        for local_variable in function.get_local_variables().values():
            new_local_variable = self.generate_local_variables(
                "nm_inline_",
                ftype=local_variable._ftype,
                shape=local_variable._shape,
                intent=local_variable.intent,
                pointer=local_variable.pointer,
                target=local_variable.target,
                allocatable=local_variable.allocatable,
                parameter=local_variable.parameter,
                assign=local_variable.assign,
                bind_c=local_variable.bind_c,
            )
            variables_couples.append((local_variable, new_local_variable))

        for stmt in function.scope.get_statements():
            Scope.add_to_current_scope(stmt.get_with_updated_variables(variables_couples))



import numpy as np
import pytest
import numeta as nm


def test_optional_argument():
    @nm.jit
    def fill(a, value=1.0):
        a[:] = value

    a = np.empty((5,), dtype=np.float64)
    fill(a)
    np.testing.assert_allclose(a, np.full((5,), 1.0, dtype=np.float64))
    fill(a, 2.0)
    np.testing.assert_allclose(a, np.full((5,), 2.0, dtype=np.float64))
    fill(a, value=3.0)
    np.testing.assert_allclose(a, np.full((5,), 3.0, dtype=np.float64))

    assert len(fill.get_symbolic_functions()) == 1


def test_optional_comptime_argument():
    @nm.jit
    def fill(a, value: nm.comptime = 1.0):
        a[:] = value

    a = np.empty((5,), dtype=np.float64)
    fill(a)
    np.testing.assert_allclose(a, np.full((5,), 1.0, dtype=np.float64))
    fill(a, 2.0)
    np.testing.assert_allclose(a, np.full((5,), 2.0, dtype=np.float64))
    fill(a, value=3.0)
    np.testing.assert_allclose(a, np.full((5,), 3.0, dtype=np.float64))

    assert len(fill.get_symbolic_functions()) == 3


def test_optional_argument_mixed():
    @nm.jit
    def fill(a, value=1.0, value2=-2.0):
        a[:] = value
        a[2] = value2

    a = np.empty((5,), dtype=np.float64)
    fill(a)
    expected = np.full((5,), 1.0, dtype=np.float64)
    expected[2] = -2.0
    np.testing.assert_allclose(a, expected)

    fill(a, 2.0)
    expected = np.full((5,), 2.0, dtype=np.float64)
    expected[2] = -2.0
    np.testing.assert_allclose(a, expected)

    fill(a, value=3.0)
    expected = np.full((5,), 3.0, dtype=np.float64)
    expected[2] = -2.0
    np.testing.assert_allclose(a, expected)

    fill(a, value2=2.0)
    expected = np.full((5,), 1.0, dtype=np.float64)
    expected[2] = 2.0
    np.testing.assert_allclose(a, expected)

    fill(a, value2=2.0, value=5.0)
    expected = np.full((5,), 5.0, dtype=np.float64)
    expected[2] = 2.0
    np.testing.assert_allclose(a, expected)

    fill(a, 2.0, 5.0)
    expected = np.full((5,), 2.0, dtype=np.float64)
    expected[2] = 5.0
    np.testing.assert_allclose(a, expected)

    assert len(fill.get_symbolic_functions()) == 1


@pytest.mark.parametrize("n_args", range(1, 5))
@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
@pytest.mark.parametrize("shape", [(), (5,), (2, 3)])
def test_keyword_arguments(n_args, dtype, shape):
    @nm.jit
    def fill(**kwargs):
        for i, arg in kwargs.items():
            arg[:] = int(i[3])

    args = {f"in_{i}": np.empty(shape, dtype=dtype) for i in range(n_args)}
    fill(**args)

    for i, arg in args.items():
        expected = np.full(shape, float(i[3]), dtype=dtype)
        np.testing.assert_allclose(arg, expected)


@pytest.mark.parametrize("n_args", range(1, 4))
@pytest.mark.parametrize("n_kwargs", range(1, 4))
def test_variable_positional_and_keyword_arguments(n_args, n_kwargs):
    @nm.jit
    def fill(*args, **kwargs):
        for i, arg in enumerate(args):
            arg[:] = int(i)
        for i, arg in kwargs.items():
            arg[:] = int(i[3])

    args = [np.empty(()) for _ in range(n_args)]
    kwargs = {f"in_{i}": np.empty(()) for i in range(n_kwargs)}
    fill(*args, **kwargs)

    for i, arg in enumerate(args):
        expected = np.full((), float(i))
        np.testing.assert_allclose(arg, expected)
    for i, arg in kwargs.items():
        expected = np.full((), float(i[3]))
        np.testing.assert_allclose(arg, expected)


@pytest.mark.parametrize("n_args", range(1, 4))
@pytest.mark.parametrize("n_kwargs", range(1, 4))
def test_positional_and_variable_positional_and_keyword_arguments(n_args, n_kwargs):
    @nm.jit
    def fill(a, *args, **kwargs):
        for i, arg in enumerate(args):
            arg[:] = int(i)
            a[:] += int(i)
        for i, arg in kwargs.items():
            arg[:] = int(i[3])
            a += int(i[3])

    args = [np.empty(()) for _ in range(n_args)]
    kwargs = {f"in_{i}": np.empty(()) for i in range(n_kwargs)}
    a = np.zeros(())
    fill(a, *args, **kwargs)

    expected_a = np.zeros(())
    for i, arg in enumerate(args):
        expected = np.full((), float(i))
        expected_a += float(i)
        np.testing.assert_allclose(arg, expected)
    for i, arg in kwargs.items():
        expected = np.full((), float(i[3]))
        expected_a += float(i[3])
        np.testing.assert_allclose(arg, expected)
    np.testing.assert_allclose(a, expected_a)


@pytest.mark.parametrize("n_args", range(1, 4))
@pytest.mark.parametrize("n_kwargs", range(1, 4))
def test_positional_and_variable_positional_and_optional_and_keyword_arguments(n_args, n_kwargs):
    @nm.jit
    def fill(a, *args, b=2.0, **kwargs):
        for i, arg in enumerate(args):
            arg[:] = int(i) + b
            a[:] += int(i)
        for i, arg in kwargs.items():
            arg[:] = int(i[3]) + b
            a += int(i[3])
        a[:] -= b

    args = [np.empty(()) for _ in range(n_args)]
    kwargs = {f"in_{i}": np.empty(()) for i in range(n_kwargs)}
    a = np.zeros(())
    fill(a, *args, **kwargs)

    expected_a = np.zeros(())
    for i, arg in enumerate(args):
        expected = np.full((), float(i) + 2.0)
        expected_a += float(i)
        np.testing.assert_allclose(arg, expected)
    for i, arg in kwargs.items():
        expected = np.full((), float(i[3]) + 2.0)
        expected_a += float(i[3])
        np.testing.assert_allclose(arg, expected)
    expected_a -= 2.0
    np.testing.assert_allclose(a, expected_a)


def test_optional_array_argument():
    @nm.jit
    def fill(a, b=np.zeros((5,), dtype=np.float64)):
        for i in nm.range(a.shape[0]):
            a[i] = b[i] + 1.0

    arr = np.empty((5,), dtype=np.float64)
    fill(arr)
    np.testing.assert_allclose(arr, np.ones((5,), dtype=np.float64))

    other = np.full((5,), 2.0, dtype=np.float64)
    fill(arr, other)
    np.testing.assert_allclose(arr, np.full((5,), 3.0, dtype=np.float64))


@pytest.mark.parametrize(
    "order",
    [
        ("a", "b", "c"),
        ("b", "c", "a"),
        ("c", "a", "b"),
    ],
)
def test_all_keyword_permutations(order):
    @nm.jit
    def fill(a, b, c=1.0):
        a[:] = b + c

    a = np.zeros(())
    params = {"a": a, "b": 2.0, "c": 3.0}
    call_kwargs = {k: params[k] for k in order}
    fill(**call_kwargs)

    np.testing.assert_allclose(a, np.full((), 5.0))


def test_optional_keyword_only_comptime_argument():

    @nm.jit
    def fill(a, *, length: nm.comptime = 4):
        for i in nm.range(length):
            a[i] = -i

    a = np.zeros(6, dtype=np.float64)
    fill(a)
    expected = np.zeros_like(a)
    expected[:4] = -np.arange(4, dtype=a.dtype)
    np.testing.assert_allclose(a, expected)

    b = np.zeros(6, dtype=np.float64)
    fill(b, length=5)
    expected = np.zeros_like(b)
    expected[:5] = -np.arange(5, dtype=b.dtype)
    np.testing.assert_allclose(b, expected)


def test_keyword_arguments_have_unique_signature():
    @nm.jit
    def fill(**kwargs):
        for i, arg in kwargs.items():
            arg[:] = float(i[3])

    args = {f"in_{i}": np.zeros(()) for i in range(3)}
    fill(**args)

    for i, arg in args.items():
        np.testing.assert_allclose(arg, float(i[3]))

    args = {f"in_{i}": np.zeros(()) for i in range(2, -1, -1)}
    fill(**args)

    for i, arg in args.items():
        np.testing.assert_allclose(arg, float(i[3]))

    assert len(fill.get_symbolic_functions()) == 1


def test_keyword_arguments_have_unique_signature_unordered():

    from numeta.settings import settings

    settings.unset_reorder_kwargs()

    @nm.jit
    def fill(**kwargs):
        for i, arg in kwargs.items():
            arg[:] = float(i[3])

    args = {f"in_{i}": np.zeros(()) for i in range(3)}
    fill(**args)

    for i, arg in args.items():
        np.testing.assert_allclose(arg, float(i[3]))

    args = {f"in_{i}": np.zeros(()) for i in range(2, -1, -1)}
    fill(**args)

    for i, arg in args.items():
        np.testing.assert_allclose(arg, float(i[3]))

    assert len(fill.get_symbolic_functions()) == 2

    settings.set_reorder_kwargs()



import numeta as nm
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_scalar_array(dtype):

    @nm.jit
    def fill(a):
        a[:] = 2

    a = np.empty(()).astype(dtype)
    fill(a)

    np.testing.assert_allclose(a, np.array([2]).astype(dtype))


def test_scalar():

    @nm.jit
    def fill(a, b):
        a[:] = b

    a = np.empty((), dtype=np.int32)
    b = np.int32(7)
    fill(a, b)

    np.testing.assert_allclose(a, np.array([7], dtype=np.int32))



import numeta as nm
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_do(dtype):
    n = 100

    @nm.jit
    def do_loop(n, a) -> None:
        i = nm.scalar(nm.i8)
        with nm.do(i, 0, n - 1):
            a[i] = i * 2

    a = np.empty(n, dtype=dtype)

    do_loop(n, a)

    if np.issubdtype(dtype, np.integer):
        np.testing.assert_allclose(a, np.array([i * 2 for i in range(n)]).astype(dtype), atol=0)
    else:
        np.testing.assert_allclose(
            a,
            np.array([i * 2 for i in range(n)]).astype(dtype),
            rtol=10e2 * np.finfo(dtype).eps,
        )



import numpy as np
import numeta as nm


def test_cond():
    @nm.jit
    def cond(a) -> None:
        a[:] = 0.0
        for i in nm.range(n):
            if nm.cond(i < 9):
                if nm.cond(i < 3):
                    if nm.cond(i < 1):
                        a[i] = 0
                    elif nm.cond(i < 2):
                        a[i] = 1
                    else:
                        a[i] = 2
                    nm.endif()
                elif nm.cond(i < 6):
                    if nm.cond(i < 4):
                        a[i] = 3
                    elif nm.cond(i < 5):
                        a[i] = 4
                    else:
                        a[i] = 5
                    nm.endif()
                else:
                    if nm.cond(i < 7):
                        a[i] = 6
                    elif nm.cond(i < 8):
                        a[i] = 7
                    else:
                        a[i] = 8
                    nm.endif()
                nm.endif()
            elif nm.cond(i < 18):
                if nm.cond(i < 12):
                    if nm.cond(i < 10):
                        a[i] = 9
                    elif nm.cond(i < 11):
                        a[i] = 10
                    else:
                        a[i] = 11
                    nm.endif()
                elif nm.cond(i < 15):
                    if nm.cond(i < 13):
                        a[i] = 12
                    elif nm.cond(i < 14):
                        a[i] = 13
                    else:
                        a[i] = 14
                    nm.endif()
                else:
                    if nm.cond(i < 16):
                        a[i] = 15
                    elif nm.cond(i < 17):
                        a[i] = 16
                    else:
                        a[i] = 17
                    nm.endif()
                nm.endif()
            else:
                if nm.cond(i < 21):
                    if nm.cond(i < 19):
                        a[i] = 18
                    elif nm.cond(i < 20):
                        a[i] = 19
                    else:
                        a[i] = 20
                    nm.endif()
                elif nm.cond(i < 24):
                    if nm.cond(i < 22):
                        a[i] = 21
                    elif nm.cond(i < 23):
                        a[i] = 22
                    else:
                        a[i] = 23
                    nm.endif()
                else:
                    if nm.cond(i < 25):
                        a[i] = 24
                    elif nm.cond(i < 26):
                        a[i] = 25
                    else:
                        a[i] = 26
                    nm.endif()
                nm.endif()
            nm.endif()

    n = 27
    a = np.empty(n, dtype=np.float64)

    cond(a)
    np.testing.assert_allclose(a, np.array(range(n), dtype=np.float64))



import numpy as np
import numeta as nm


def test_inline_array_scalar():

    @nm.jit(inline=True)
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(n, a)

    a = np.zeros((), dtype=np.int64)
    caller(1, a)

    expected = np.zeros((), dtype=np.int64)
    expected[...] = 1
    np.testing.assert_equal(a, expected)


def test_inline_array():

    @nm.jit(inline=True)
    def callee(n, a):
        a[:, 2] = n

    @nm.jit
    def caller(n, a):
        callee(n, a)

    a = np.zeros((5, 10), dtype=np.int64)
    caller(1, a)

    expected = np.zeros((5, 10), dtype=np.int64)
    expected[:, 2] = 1
    np.testing.assert_equal(a, expected)


def test_inline_getitem_scalar():

    @nm.jit(inline=True)
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(n, a[3, 7])

    a = np.zeros((5, 10), dtype=np.int64)
    caller(1, a)

    expected = np.zeros((5, 10), dtype=np.int64)
    expected[3, 7] = 1
    np.testing.assert_equal(a, expected)


def test_inline_getitem_slice():

    @nm.jit(inline=True)
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(50, a[:])
        callee(n, a[:2, :3])
        callee(n + 1, a[3:4, 8:])
        callee(n + 2, a[:, 7])

    a = np.zeros((5, 10), dtype=np.int64)
    caller(1, a)

    expected = np.zeros((5, 10), dtype=np.int64)
    expected[:] = 50
    expected[:2, :3] = 1
    expected[3:4, 8:] = 2
    expected[:, 7] = 3
    np.testing.assert_equal(a, expected)


def test_inline_getitem_slice_runtime_dep():

    @nm.jit(inline=True)
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, m, a):
        callee(50, a[:])
        callee(n, a[:n, : m // 2])
        callee(2, a[2 : n + 1, m - 2 :])
        callee(3, a[:, n])
        callee(1, a[:, n : m - n])

    n = 4
    m = 9
    a = np.zeros((5, 10), dtype=np.int64)
    caller(n, m, a)

    expected = np.zeros((5, 10), dtype=np.int64)
    expected[:] = 50
    expected[:n, : m // 2] = n
    expected[2 : n + 1, m - 2 :] = 2
    expected[:, n] = 3
    expected[:, n : m - n] = 1
    np.testing.assert_equal(a, expected)


def test_inline_getattr_scalar():

    @nm.jit(inline=True)
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(n, a["x"])

    dtype = np.dtype([("x", np.int64), ("y", np.float64, (2, 2))], align=True)

    a = np.zeros((), dtype=dtype)
    caller(1, a)

    expected = np.zeros((), dtype=dtype)
    expected["x"] = 1
    np.testing.assert_equal(a, expected)


def test_inline_getattr_array():

    @nm.jit(inline=True)
    def callee(n, a):
        a[:] = n

    @nm.jit
    def caller(n, a):
        callee(n, a["y"])

    dtype = np.dtype([("x", np.int64), ("y", np.float64, (2, 2))], align=True)

    a = np.zeros((), dtype=dtype)
    caller(2.0, a)

    expected = np.zeros((), dtype=dtype)
    expected["y"] = 2.0
    np.testing.assert_equal(a, expected)


def test_inline_matmul():

    @nm.jit(inline=True)
    def callee(a, d):
        a[:] = d

    @nm.jit
    def caller(a, b, c):
        callee(c, nm.matmul(b, a))

    n = 10
    a = np.random.random((n, n))
    b = np.random.random((n, n))
    c = np.zeros((10, 10))
    caller(a, b, c)

    expected = a @ b
    np.testing.assert_equal(c, expected)


def test_inline_nested_calls():
    @nm.jit(inline=True)
    def inner(n, arr):
        arr[0] = n

    @nm.jit(inline=True)
    def middle(n, arr):
        inner(n, arr)
        arr[1] = n + 1

    @nm.jit
    def caller(n, arr):
        middle(n, arr)

    arr = np.zeros(2, dtype=np.int64)
    caller(5, arr)

    expected = np.array([5, 6], dtype=np.int64)
    np.testing.assert_equal(arr, expected)


def test_inline_nested_calls_dependencies():
    """Test to check that inline nested calls propagate dependencies correctly."""

    @nm.jit
    def inner(n, arr):
        arr[0] = n

    @nm.jit(inline=True)
    def middle(n, arr):
        inner(n, arr)
        arr[1] = n + 1

    @nm.jit
    def caller(n, arr):
        middle(n, arr)

    arr = np.zeros(2, dtype=np.int64)
    caller(5, arr)

    expected = np.array([5, 6], dtype=np.int64)
    np.testing.assert_equal(arr, expected)


def test_inline_matmul_fortran_order():

    @nm.jit(inline=True)
    def callee(a, d):
        a[:] = d

    @nm.jit
    def caller(a, b, c):
        callee(c, nm.matmul(a, b))

    n = 10
    a = np.random.random((n, n)).astype(np.float64, order="F")
    b = np.random.random((n, n)).astype(np.float64, order="F")
    c = np.zeros((10, 10), order="F")
    caller(a, b, c)

    expected = a @ b
    np.testing.assert_equal(c, expected)


def test_inline_loop_matmul():
    @nm.jit(inline=True)
    def callee(n, a, b, c):
        for i in nm.range(n):
            for k in nm.range(n):
                c[i, :] += a[i, k] * b[k, :]

    @nm.jit
    def caller(n, a, b, c):
        callee(n, a, b, c)

    n = 4
    a = np.random.random((n, n))
    b = np.random.random((n, n))
    c = np.zeros((n, n))
    caller(n, a, b, c)

    expected = a @ b
    np.testing.assert_allclose(c, expected)


def test_inline_loop_if():
    @nm.jit(inline=True)
    def callee(n, arr):
        for i in nm.range(n):
            with nm.If(i < 2):
                arr[i] = i
            with nm.ElseIf(i < n // 2):
                arr[i] = i + 1
            with nm.Else():
                arr[i] = -i

    @nm.jit
    def caller(n, arr):
        callee(n, arr)

    n = 10
    arr = np.zeros(n, dtype=np.int64)
    caller(n, arr)

    expected = np.array([0, 1, 3, 4, 5, -5, -6, -7, -8, -9], dtype=np.int64)
    np.testing.assert_equal(arr, expected)


def test_inline_jit_threshold_inline():
    @nm.jit(inline=2)
    def callee(a):
        a[:] = 1
        a[:] = 2

    @nm.jit
    def caller(a):
        callee(a)

    arr = np.zeros(5, dtype=np.int64)
    caller(arr)

    expected = np.full(5, 2, dtype=np.int64)
    np.testing.assert_equal(arr, expected)

    signature = caller.get_signature(arr)
    assert len(caller._compiled_targets[signature].get_nested_obj_files()) == 0


def test_inline_jit_threshold_call():
    @nm.jit(inline=1)
    def callee(a):
        a[:] = 1
        a[:] = 2

    @nm.jit
    def caller(a):
        callee(a)

    arr = np.zeros(5, dtype=np.int64)
    caller(arr)

    expected = np.full(5, 2, dtype=np.int64)
    np.testing.assert_equal(arr, expected)

    signature = caller.get_signature(arr)
    assert len(caller._compiled_targets[signature].get_nested_obj_files()) != 0


def test_inline_jit_threshold_loop_inline():
    @nm.jit(inline=3)
    def callee(n, a):
        for i in nm.range(n):
            for j in nm.range(n):
                a[i, j] += 1

    @nm.jit
    def caller(n, a):
        callee(n, a)

    arr = np.zeros((3, 3), dtype=np.int64)
    caller(3, arr)

    expected = np.ones((3, 3), dtype=np.int64)
    np.testing.assert_equal(arr, expected)

    signature = caller.get_signature(3, arr)
    assert len(caller._compiled_targets[signature].get_nested_obj_files()) == 0


def test_inline_jit_threshold_loop_call():
    @nm.jit(inline=2)
    def callee(n, a):
        for i in nm.range(n):
            for j in nm.range(n):
                a[i, j] += 1

    @nm.jit
    def caller(n, a):
        callee(n, a)

    arr = np.zeros((3, 3), dtype=np.int64)
    caller(3, arr)

    expected = np.ones((3, 3), dtype=np.int64)
    np.testing.assert_equal(arr, expected)

    signature = caller.get_signature(3, arr)
    assert len(caller._compiled_targets[signature].get_nested_obj_files()) != 0


def test_inline_jit_threshold_if_inline():
    @nm.jit(inline=5)
    def callee(n, a):
        for i in nm.range(n):
            with nm.If(i < 2):
                a[i] = i
            with nm.Else():
                a[i] = -i

    @nm.jit
    def caller(n, a):
        callee(n, a)

    arr = np.zeros(5, dtype=np.int64)
    caller(5, arr)

    expected = np.array([0, 1, -2, -3, -4], dtype=np.int64)
    np.testing.assert_equal(arr, expected)

    signature = caller.get_signature(5, arr)
    assert len(caller._compiled_targets[signature].get_nested_obj_files()) == 0


def test_inline_jit_threshold_if_call():
    @nm.jit(inline=4)
    def callee(n, a):
        for i in nm.range(n):
            with nm.If(i < 2):
                a[i] = i
            with nm.Else():
                a[i] = -i

    @nm.jit
    def caller(n, a):
        callee(n, a)

    arr = np.zeros(5, dtype=np.int64)
    caller(5, arr)

    expected = np.array([0, 1, -2, -3, -4], dtype=np.int64)
    np.testing.assert_equal(arr, expected)

    signature = caller.get_signature(5, arr)
    assert len(caller._compiled_targets[signature].get_nested_obj_files()) != 0


def test_inline_tmp_scalar():

    @nm.jit(inline=True)
    def callee(n, a):
        i = nm.int64(5)
        a[:] = n + i

    @nm.jit
    def caller(a):
        n = nm.int64(1, name="n")
        callee(n, a)

    a = np.zeros((), dtype=np.int64)
    caller(a)

    expected = np.zeros((), dtype=np.int64)
    expected[...] = 6
    np.testing.assert_equal(a, expected)


def test_inline_name_mangling():
    """
    To test if the local variables are properly renamed.
    """

    @nm.jit(inline=True)
    def callee(n, a):
        f = nm.float64(5.0)
        a[n] = f

    @nm.jit(directory="build")
    def caller(a):
        n = nm.int32(6)
        callee(n, a)

    a = np.zeros(10, dtype=np.int64)
    caller(a)

    expected = np.zeros(10, dtype=np.int64)
    expected[6] = 5.0
    np.testing.assert_equal(a, expected)


def test_inline_slice_composition():

    @nm.jit(inline=True)
    def callee(arr):
        arr[:4] = 1.0
        arr[4:6] = 2.0
        arr[6:] = 3.0
        arr[2] = 4.0

    @nm.jit
    def caller(a, b, c, d):
        callee(a[3:10])
        callee(b)
        callee(c[5:])
        callee(d[:10])

    a = np.zeros(12)
    b = np.zeros(12)
    c = np.zeros(12)
    d = np.zeros(12)
    caller(a, b, c, d)

    expected_a = np.zeros(12)
    expected_a[3:7] = 1.0
    expected_a[7:9] = 2.0
    expected_a[9:10] = 3.0
    expected_a[5] = 4.0
    np.testing.assert_allclose(a, expected_a)

    expected_b = np.zeros(12)
    expected_b[:7] = 1.0
    expected_b[4:6] = 2.0
    expected_b[6:] = 3.0
    expected_b[2] = 4.0
    np.testing.assert_allclose(b, expected_b)

    expected_c = np.zeros(12)
    expected_c[5:12] = 1.0
    expected_c[9:11] = 2.0
    expected_c[11:] = 3.0
    expected_c[7] = 4.0
    np.testing.assert_allclose(c, expected_c)
    expected_c = np.zeros(12)
    expected_c[5:12] = 1.0
    expected_c[9:11] = 2.0
    expected_c[11:] = 3.0
    expected_c[7] = 4.0
    np.testing.assert_allclose(c, expected_c)

    expected_d = np.zeros(12)
    expected_d[:7] = 1.0
    expected_d[4:6] = 2.0
    expected_d[6:10] = 3.0
    expected_d[2] = 4.0
    np.testing.assert_allclose(d, expected_d)



from numeta.syntax.nodes import Node
from abc import abstractmethod


class ExpressionNode(Node):
    __slots__ = []

    def __init__(self):
        pass

    @property
    @abstractmethod
    def _ftype(self):
        """Return the Fortran type of the expression."""

    @property
    @abstractmethod
    def _shape(self):
        """Return the shape of the expression if any."""

    def _get_shape_descriptor(self):
        from .various import ArrayConstructor

        return ArrayConstructor(*self._shape.dims)

    @abstractmethod
    def extract_entities(self):
        """Extract the nested entities of the expression."""

    @abstractmethod
    def get_code_blocks(self):
        """Return the code blocks that represent the expression."""

    def get_with_updated_variables(self, variables_couples):
        return self

    def __bool__(self) -> bool:
        raise Warning("Do not use 'bool' operator for expressions.")

    def __rshift__(self, other):
        from numeta.syntax.statements import Assignment

        if isinstance(other, (int, float, complex, bool, str)):
            from .literal_node import LiteralNode

            other = LiteralNode(other)
        return Assignment(self, other)

    def __neg__(self):
        from .intrinsic_functions import Neg

        return Neg(self)

    def __abs__(self):
        from .intrinsic_functions import Abs

        return Abs(self)

    def __add__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, "+", other)

    def __radd__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(other, "+", self)

    def __sub__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, "-", other)

    def __rsub__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(other, "-", self)

    def __mul__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, "*", other)

    def __rmul__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(other, "*", self)

    def __truediv__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, "/", other)

    def __rtruediv__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(other, "/", self)

    def __floordiv__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, "/", other)

    def __rfloordiv__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(other, "/", self)

    def __pow__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, "**", other)

    def __rpow__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(other, "**", self)

    def __and__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, ".and.", other)

    def __or__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, ".or.", other)

    def __ne__(self, other):
        from .binary_operation_node import NeBinaryNode

        return NeBinaryNode(self, other)

    def __eq__(self, other):
        from .binary_operation_node import EqBinaryNode

        return EqBinaryNode(self, other)

    def __ge__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, ".ge.", other)

    def __gt__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, ".gt.", other)

    def __le__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, ".le.", other)

    def __lt__(self, other):
        from .binary_operation_node import BinaryOperationNode

        return BinaryOperationNode(self, ".lt.", other)

    @property
    def real(self):
        from .various import Re

        return Re(self)

    @property
    def imag(self):
        from .various import Im

        return Im(self)

    @property
    def T(self):
        from .intrinsic_functions import Transpose

        return Transpose(self)

    def __getitem__(self, key):
        if isinstance(key, slice) and key.start is None and key.stop is None and key.step is None:
            return self
        from .getitem import GetItem
        from .getattr import GetAttr

        if isinstance(key, str):
            return GetAttr(self, key)
        return GetItem(self, key)



import numpy as np
from numeta.syntax import Variable, Module
from numeta.datatype import DataType, float64, FortranType, get_datatype
from numeta.array_shape import ArrayShape
from numeta.numeta_function import NumetaCompilationTarget

_n_global_constant = 0


def declare_global_constant(
    shape,
    dtype: DataType | FortranType | np.generic = float64,
    order="C",
    name=None,
    value=None,
    directory=None,
):
    if order not in ["C", "F"]:
        raise ValueError(f"Invalid order: {order}, must be 'C' or 'F'")

    fortran_order = order == "F"
    if not isinstance(shape, ArrayShape):
        if not isinstance(shape, (tuple, list)):
            shape = (shape,)
        shape = ArrayShape(shape, fortran_order=fortran_order)

    if isinstance(dtype, FortranType):
        ftype = dtype
    else:
        ftype = get_datatype(dtype).get_fortran()

    if name is None:
        global _n_global_constant
        name = f"global_constant_{_n_global_constant}"
        _n_global_constant += 1

    # Lets create a module to host the global_constant variable, it will be a module variable
    global_constant_var_module = Module(f"{name}_module")

    var = Variable(
        name=name,
        ftype=ftype,
        shape=shape,
        assign=value,
        # TODO
        # parameter=True, # parameter is not supported yet, so not really constant.
        parent=global_constant_var_module,
    )

    # We have to compile the module when needed
    module_library = NumetaCompilationTarget(
        f"{name}_module",
        global_constant_var_module,
        directory=directory,
    )
    global_constant_var_module.parent = module_library
    return var



from .nodes import Node, NamedEntity


class FortranType(Node):
    def __init__(self, type_, kind) -> None:
        super().__init__()
        self.type = type_
        self.kind = kind
        self.parent = None

    def get_code_blocks(self):
        return [self.type, "(", self.get_kind_str(), ")"]

    def extract_entities(self):
        if isinstance(self.kind, NamedEntity):
            yield self.kind

    def get_with_updated_variables(self, variables_couples):
        return self

    def get_kind_str(self):
        if isinstance(self.kind, NamedEntity):
            return self.kind.name
        return str(self.kind)



import numpy as np
import numbers


def extract_entities(element):
    """Yield entities referenced by ``element``, recursively walking containers and slices."""
    if hasattr(element, "extract_entities"):
        yield from element.extract_entities()
    elif isinstance(element, (tuple, list)):
        for e in element:
            yield from extract_entities(e)
    elif isinstance(element, slice):
        yield from extract_entities(element.start)
        yield from extract_entities(element.stop)
        yield from extract_entities(element.step)


def check_node(node):
    """Return a LiteralNode for scalars, otherwise pass through existing nodes."""
    if isinstance(node, (numbers.Number, np.generic, bool, str)):
        from .expressions import LiteralNode

        return LiteralNode(node)
    else:
        return node
        # otherwise is so slow
        # TODO: maybe to move check node at the print time
        from .nodes import Node

        if isinstance(node, Node):
            return node
        else:
            raise ValueError(f"Unknown node type: {node.__class__.__name__} value: {node}")


def update_variables(element, variables_couples):
    """Recursively replace variables in ``element`` using ``variables_couples``."""
    if isinstance(element, tuple):
        return tuple(update_variables(e, variables_couples) for e in element)
    if isinstance(element, list):
        return [update_variables(e, variables_couples) for e in element]
    if isinstance(element, slice):
        return slice(
            update_variables(element.start, variables_couples),
            update_variables(element.stop, variables_couples),
            update_variables(element.step, variables_couples),
        )
    from numeta.array_shape import ArrayShape

    if isinstance(element, ArrayShape):
        return ArrayShape(
            tuple(update_variables(dim, variables_couples) for dim in element.dims),
            fortran_order=element.fortran_order,
        )
    from numeta.syntax.nodes.base_node import Node

    if isinstance(element, Node):
        return element.get_with_updated_variables(variables_couples)
    return element



class Scope:
    current_scope = None

    @classmethod
    def add_to_current_scope(cls, statement):
        if cls.current_scope is None:
            raise Exception("No current scope")
        cls.current_scope.__body.append(statement)

    @classmethod
    def end(cls):
        if not issubclass(cls, Scope):
            raise Exception("end() must be called from a subclass of Scope")
        if cls.current_scope is None:
            raise Exception("No current scope")
        cls.current_scope = cls.current_scope.parent

    def __init__(self, parent=None) -> None:
        self.parent = parent if parent is not None else Scope.current_scope
        self.__body = []

    @property
    def body(self):
        return self.__body

    @body.setter
    def body(self, value):
        self.__body = value

    def enter(self):
        Scope.current_scope = self

    def exit(self):
        Scope.current_scope = self.parent

    def get_statements(self):
        return self.body

    def get_with_updated_variables(self, variables_couples):
        result = Scope()
        result.body = [
            stmt.get_with_updated_variables(variables_couples) for stmt in self.get_statements()
        ]
        return result

    def __enter__(self):
        self.enter()

    def __exit__(self, *args):
        self.exit()



from .numpy_mem import numpy_mem
from .range import range
from .prange import prange
from .cond import cond, endif
from .scalar import scalar
from .cases import cases
from .empty import empty
from .zeros import zeros
from .reshape import reshape
from .external_library import ExternalLibraryWrapper
from .declare_global_constant import declare_global_constant
from .constant import constant
from .cast import cast



import numeta as nm
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_zeros(dtype):
    n = 50
    m = 20

    @nm.jit
    def copy_and_set_zero_first_col_with_zeros(a):
        tmp = nm.zeros((n, m), dtype)

        for i in nm.range(n):
            for j in nm.range(m):
                a[i, j] = tmp[i, j]

    a = np.ones((n, m)).astype(dtype)
    copy_and_set_zero_first_col_with_zeros(a)

    c = np.zeros((n, m)).astype(dtype)

    np.testing.assert_allclose(a, c)


@pytest.mark.parametrize(
    "dtype", [np.float64, np.float32, np.int64, np.int32, np.complex64, np.complex128]
)
def test_zeros_fortran(dtype):
    n = 50
    m = 20

    @nm.jit
    def copy_and_set_zero_first_col_with_zeros(a):
        tmp = nm.zeros((n, m), dtype, order="F")

        for i in nm.range(n):
            for j in nm.range(m):
                a[i, j] = tmp[i, j]

    a = np.ones((n, m)).astype(dtype)
    copy_and_set_zero_first_col_with_zeros(a)

    c = np.zeros((n, m)).astype(dtype)

    np.testing.assert_allclose(a, c)



from .expression_node import ExpressionNode
from numeta.syntax.tools import check_node
from numeta.array_shape import ArrayShape
from numeta.syntax.settings import settings


class Re(ExpressionNode):
    def __init__(self, variable):
        self.variable = variable

    @property
    def _ftype(self):
        return settings.DEFAULT_REAL

    @property
    def _shape(self):
        return self.variable._shape

    def get_code_blocks(self):
        return [*self.variable.get_code_blocks(), "%", "re"]

    def extract_entities(self):
        yield from self.variable.extract_entities()

    def get_with_updated_variables(self, variables_couples):
        return Re(self.variable.get_with_updated_variables(variables_couples))


class Im(ExpressionNode):
    def __init__(self, variable):
        self.variable = variable

    @property
    def _ftype(self):
        return settings.DEFAULT_REAL

    @property
    def _shape(self):
        return self.variable.shape

    def get_code_blocks(self):
        return [*self.variable.get_code_blocks(), "%", "im"]

    def extract_entities(self):
        yield from self.variable.extract_entities()

    def get_with_updated_variables(self, variables_couples):
        return Im(self.variable.get_with_updated_variables(variables_couples))


class ArrayConstructor(ExpressionNode):
    def __init__(self, *elements):
        self.elements = [check_node(e) for e in elements]

    @property
    def _ftype(self):
        if not self.elements:
            raise ValueError("ArrayConstructor must have at least one element")
        return self.elements[0].dtype

    @property
    def _shape(self):
        return ArrayShape((len(self.elements),))

    def get_code_blocks(self):
        result = ["["]
        for element in self.elements:
            if element is None:
                result.append("None")
            else:
                result.extend(element.get_code_blocks())
            result.append(", ")
        if result[-1] == ", ":
            result.pop()
        result.append("]")
        return result

    def extract_entities(self):
        for e in self.elements:
            if e is not None:
                yield from e.extract_entities()

    def get_with_updated_variables(self, variables_couples):
        new_elements = [e.get_with_updated_variables(variables_couples) for e in self.elements]
        return ArrayConstructor(*new_elements)
